# Copyright (C) 2006-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test for 'brz mv'."""

import os

import breezy.branch
from breezy import osutils, workingtree
from breezy.tests import TestCaseWithTransport, script
from breezy.tests.features import (
    CaseInsensitiveFilesystemFeature,
    SymlinkFeature,
    UnicodeFilenameFeature,
)


class TestMove(TestCaseWithTransport):
    def assertMoved(self, from_path, to_path):
        """Assert that to_path is existing and versioned but from_path not."""
        self.assertPathDoesNotExist(from_path)
        self.assertNotInWorkingTree(from_path)

        self.assertPathExists(to_path)
        self.assertInWorkingTree(to_path)

    def test_mv_modes(self):
        """Test two modes of operation for mv."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "c", "subdir/"])
        tree.add(["a", "c", "subdir"])

        self.run_bzr("mv a b")
        self.assertMoved("a", "b")

        self.run_bzr("mv b subdir")
        self.assertMoved("b", "subdir/b")

        self.run_bzr("mv subdir/b a")
        self.assertMoved("subdir/b", "a")

        self.run_bzr("mv a c subdir")
        self.assertMoved("a", "subdir/a")
        self.assertMoved("c", "subdir/c")

        self.run_bzr("mv subdir/a subdir/newa")
        self.assertMoved("subdir/a", "subdir/newa")

    def test_mv_unversioned(self):
        self.build_tree(["unversioned.txt"])
        self.run_bzr_error(
            [
                "^brz: ERROR: Could not rename unversioned.txt => elsewhere."
                " .*unversioned.txt is not versioned\\.$"
            ],
            "mv unversioned.txt elsewhere",
        )

    def test_mv_nonexisting(self):
        self.run_bzr_error(
            [
                "^brz: ERROR: Could not rename doesnotexist => somewhereelse."
                " .*doesnotexist is not versioned\\.$"
            ],
            "mv doesnotexist somewhereelse",
        )

    def test_mv_unqualified(self):
        self.run_bzr_error(["^brz: ERROR: missing file argument$"], "mv")

    def test_mv_invalid(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["test.txt", "sub1/"])
        tree.add(["test.txt"])

        self.run_bzr_error(
            ["^brz: ERROR: Could not move to sub1: sub1 is not versioned\\.$"],
            "mv test.txt sub1",
        )

        self.run_bzr_error(
            [
                "^brz: ERROR: Could not move test.txt => .*hello.txt: "
                "sub1 is not versioned\\.$"
            ],
            "mv test.txt sub1/hello.txt",
        )

    def test_mv_dirs(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["hello.txt", "sub1/"])
        tree.add(["hello.txt", "sub1"])

        self.run_bzr("mv sub1 sub2")
        self.assertMoved("sub1", "sub2")

        self.run_bzr("mv hello.txt sub2")
        self.assertMoved("hello.txt", "sub2/hello.txt")

        self.build_tree(["sub1/"])
        tree.add(["sub1"])
        self.run_bzr("mv sub2/hello.txt sub1")
        self.assertMoved("sub2/hello.txt", "sub1/hello.txt")

        self.run_bzr("mv sub2 sub1")
        self.assertMoved("sub2", "sub1/sub2")

    def test_mv_relative(self):
        self.build_tree(["sub1/", "sub1/sub2/", "sub1/hello.txt"])
        tree = self.make_branch_and_tree(".")
        tree.add(["sub1", "sub1/sub2", "sub1/hello.txt"])

        self.run_bzr("mv ../hello.txt .", working_dir="sub1/sub2")
        self.assertPathExists("sub1/sub2/hello.txt")

        self.run_bzr("mv sub2/hello.txt .", working_dir="sub1")
        self.assertMoved("sub1/sub2/hello.txt", "sub1/hello.txt")

    def test_mv_change_case_file(self):
        # test for bug #77740 (mv unable change filename case on Windows)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["test.txt"])
        tree.add(["test.txt"])
        self.run_bzr("mv test.txt Test.txt")
        # we can't use failUnlessExists on case-insensitive filesystem
        # so try to check shape of the tree
        shape = sorted(os.listdir("."))
        self.assertEqual([".bzr", "Test.txt"], shape)
        self.assertInWorkingTree("Test.txt")
        self.assertNotInWorkingTree("test.txt")

    def test_mv_change_case_dir(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo/"])
        tree.add(["foo"])
        self.run_bzr("mv foo Foo")
        # we can't use failUnlessExists on case-insensitive filesystem
        # so try to check shape of the tree
        shape = sorted(os.listdir("."))
        self.assertEqual([".bzr", "Foo"], shape)
        self.assertInWorkingTree("Foo")
        self.assertNotInWorkingTree("foo")

    def test_mv_change_case_dir_w_files(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo/", "foo/bar"])
        tree.add(["foo"])
        self.run_bzr("mv foo Foo")
        # we can't use failUnlessExists on case-insensitive filesystem
        # so try to check shape of the tree
        shape = sorted(os.listdir("."))
        self.assertEqual([".bzr", "Foo"], shape)
        self.assertInWorkingTree("Foo")
        self.assertNotInWorkingTree("foo")

    def test_mv_file_to_wrong_case_dir(self):
        self.requireFeature(CaseInsensitiveFilesystemFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["foo/", "bar"])
        tree.add(["foo", "bar"])
        out, err = self.run_bzr("mv bar Foo", retcode=3)
        self.assertEqual("", out)
        self.assertEqual(
            "brz: ERROR: Could not move to Foo: Foo is not versioned.\n", err
        )

    def test_mv_smoke_aliases(self):
        # just test that aliases for mv exist, if their behaviour is changed in
        # the future, then extend the tests.
        self.build_tree(["a"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a"])

        self.run_bzr("move a b")
        self.run_bzr("rename b a")

    def test_mv_no_root(self):
        self.make_branch_and_tree(".")
        self.run_bzr_error(["brz: ERROR: can not move root of branch"], "mv . a")

    def test_mv_through_symlinks(self):
        self.requireFeature(SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b"])
        os.symlink("a", "c")
        os.symlink(".", "d")
        tree.add(["a", "a/b", "c"], ids=[b"a-id", b"b-id", b"c-id"])
        self.run_bzr("mv c/b b")
        tree = workingtree.WorkingTree.open(".")
        self.assertEqual(b"b-id", tree.path2id("b"))

    def test_mv_already_moved_file(self):
        """Test brz mv original_file to moved_file.

        Tests if a file which has allready been moved by an external tool,
        is handled correctly by brz mv.
        Setup: a is in the working tree, b does not exist.
        User does: mv a b; brz mv a b
        """
        self.build_tree(["a"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a"])

        osutils.rename("a", "b")
        self.run_bzr("mv a b")
        self.assertMoved("a", "b")

    def test_mv_already_moved_file_to_versioned_target(self):
        """Test brz mv existing_file to versioned_file.

        Tests if an attempt to move an existing versioned file
        to another versiond file will fail.
        Setup: a and b are in the working tree.
        User does: rm b; mv a b; brz mv a b
        """
        self.build_tree(["a", "b"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a", "b"])

        os.remove("b")
        osutils.rename("a", "b")
        self.run_bzr_error(
            ["^brz: ERROR: Could not move a => b. b is already versioned\\.$"], "mv a b"
        )
        # check that nothing changed
        self.assertPathDoesNotExist("a")
        self.assertPathExists("b")

    def test_mv_already_moved_file_into_subdir(self):
        """Test brz mv original_file to versioned_directory/file.

        Tests if a file which has already been moved into a versioned
        directory by an external tool, is handled correctly by brz mv.
        Setup: a and sub/ are in the working tree.
        User does: mv a sub/a; brz mv a sub/a
        """
        self.build_tree(["a", "sub/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a", "sub"])

        osutils.rename("a", "sub/a")
        self.run_bzr("mv a sub/a")
        self.assertMoved("a", "sub/a")

    def test_mv_already_moved_file_into_unversioned_subdir(self):
        """Test brz mv original_file to unversioned_directory/file.

        Tests if an attempt to move an existing versioned file
        into an unversioned directory will fail.
        Setup: a is in the working tree, sub/ is not.
        User does: mv a sub/a; brz mv a sub/a
        """
        self.build_tree(["a", "sub/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a"])

        osutils.rename("a", "sub/a")
        self.run_bzr_error(
            ["^brz: ERROR: Could not move a => a: sub is not versioned\\.$"],
            "mv a sub/a",
        )
        self.assertPathDoesNotExist("a")
        self.assertPathExists("sub/a")

    def test_mv_already_moved_files_into_subdir(self):
        """Test brz mv original_files to versioned_directory.

        Tests if files which has already been moved into a versioned
        directory by an external tool, is handled correctly by brz mv.
        Setup: a1, a2, sub are in the working tree.
        User does: mv a1 sub/.; brz mv a1 a2 sub
        """
        self.build_tree(["a1", "a2", "sub/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a1", "a2", "sub"])

        osutils.rename("a1", "sub/a1")
        self.run_bzr("mv a1 a2 sub")
        self.assertMoved("a1", "sub/a1")
        self.assertMoved("a2", "sub/a2")

    def test_mv_already_moved_files_into_unversioned_subdir(self):
        """Test brz mv original_file to unversioned_directory.

        Tests if an attempt to move existing versioned file
        into an unversioned directory will fail.
        Setup: a1, a2 are in the working tree, sub is not.
        User does: mv a1 sub/.; brz mv a1 a2 sub
        """
        self.build_tree(["a1", "a2", "sub/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a1", "a2"])

        osutils.rename("a1", "sub/a1")
        self.run_bzr_error(
            ["^brz: ERROR: Could not move to sub. sub is not versioned\\.$"],
            "mv a1 a2 sub",
        )
        self.assertPathDoesNotExist("a1")
        self.assertPathExists("sub/a1")
        self.assertPathExists("a2")
        self.assertPathDoesNotExist("sub/a2")

    def test_mv_already_moved_file_forcing_after(self):
        """Test brz mv versioned_file to unversioned_file.

        Tests if an attempt to move an existing versioned file to an existing
        unversioned file will fail, informing the user to use the --after
        option to force this.
        Setup: a is in the working tree, b not versioned.
        User does: mv a b; touch a; brz mv a b
        """
        self.build_tree(["a", "b"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a"])

        osutils.rename("a", "b")
        self.build_tree(["a"])  # touch a
        self.run_bzr_error(
            [
                "^brz: ERROR: Could not rename a => b because both files exist."
                " \\(Use --after to tell brz about a rename that has already"
                " happened\\)$"
            ],
            "mv a b",
        )
        self.assertPathExists("a")
        self.assertPathExists("b")

    def test_mv_already_moved_file_using_after(self):
        """Test brz mv --after versioned_file to unversioned_file.

        Tests if an existing versioned file can be forced to move to an
        existing unversioned file using the --after option. With the result
        that bazaar considers the unversioned_file to be moved from
        versioned_file and versioned_file will become unversioned.
        Setup: a is in the working tree and b exists.
        User does: mv a b; touch a; brz mv a b --after
        Resulting in a => b and a is unknown.
        """
        self.build_tree(["a", "b"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a"])
        osutils.rename("a", "b")
        self.build_tree(["a"])  # touch a

        self.run_bzr("mv a b --after")
        self.assertPathExists("a")
        self.assertNotInWorkingTree("a")  # a should be unknown now.
        self.assertPathExists("b")
        self.assertInWorkingTree("b")

    def test_mv_already_moved_files_forcing_after(self):
        """Test brz mv versioned_files to directory/unversioned_file.

        Tests if an attempt to move an existing versioned file to an existing
        unversioned file in some other directory will fail, informing the user
        to use the --after option to force this.

        Setup: a1, a2, sub are versioned and in the working tree,
               sub/a1, sub/a2 are in working tree.
        User does: mv a* sub; touch a1; touch a2; brz mv a1 a2 sub
        """
        self.build_tree(["a1", "a2", "sub/", "sub/a1", "sub/a2"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a1", "a2", "sub"])
        osutils.rename("a1", "sub/a1")
        osutils.rename("a2", "sub/a2")
        self.build_tree(["a1"])  # touch a1
        self.build_tree(["a2"])  # touch a2

        self.run_bzr_error(
            [
                "^brz: ERROR: Could not rename a1 => sub/a1 because both files"
                " exist. \\(Use --after to tell brz about a rename that has already"
                " happened\\)$"
            ],
            "mv a1 a2 sub",
        )
        self.assertPathExists("a1")
        self.assertPathExists("a2")
        self.assertPathExists("sub/a1")
        self.assertPathExists("sub/a2")

    def test_mv_already_moved_files_using_after(self):
        """Test brz mv --after versioned_file to directory/unversioned_file.

        Tests if an existing versioned file can be forced to move to an
        existing unversioned file in some other directory using the --after
        option. With the result that bazaar considers
        directory/unversioned_file to be moved from versioned_file and
        versioned_file will become unversioned.

        Setup: a1, a2, sub are versioned and in the working tree,
               sub/a1, sub/a2 are in working tree.
        User does: mv a* sub; touch a1; touch a2; brz mv a1 a2 sub --after
        """
        self.build_tree(["a1", "a2", "sub/", "sub/a1", "sub/a2"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a1", "a2", "sub"])
        osutils.rename("a1", "sub/a1")
        osutils.rename("a2", "sub/a2")
        self.build_tree(["a1"])  # touch a1
        self.build_tree(["a2"])  # touch a2

        self.run_bzr("mv a1 a2 sub --after")
        self.assertPathExists("a1")
        self.assertPathExists("a2")
        self.assertPathExists("sub/a1")
        self.assertPathExists("sub/a2")
        self.assertInWorkingTree("sub/a1")
        self.assertInWorkingTree("sub/a2")

    def test_mv_already_moved_directory(self):
        """Use `brz mv a b` to mark a directory as renamed.

        https://bugs.launchpad.net/bzr/+bug/107967/
        """
        self.build_tree(["a/", "c/"])
        tree = self.make_branch_and_tree(".")
        tree.add(["a", "c"])
        osutils.rename("a", "b")
        osutils.rename("c", "d")
        # mv a b should work just like it does for already renamed files
        self.run_bzr("mv a b")
        self.assertPathDoesNotExist("a")
        self.assertNotInWorkingTree("a")
        self.assertPathExists("b")
        self.assertInWorkingTree("b")
        # and --after should work, too (technically it's ignored)
        self.run_bzr("mv --after c d")
        self.assertPathDoesNotExist("c")
        self.assertNotInWorkingTree("c")
        self.assertPathExists("d")
        self.assertInWorkingTree("d")

    def make_abcd_tree(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a", "tree/c"])
        tree.add(["a", "c"])
        tree.commit("record old names")
        osutils.rename("tree/a", "tree/b")
        osutils.rename("tree/c", "tree/d")
        return tree

    def test_mv_auto(self):
        self.make_abcd_tree()
        out, err = self.run_bzr("mv --auto", working_dir="tree")
        self.assertEqual(out, "")
        self.assertEqual(err, "a => b\nc => d\n")
        tree = workingtree.WorkingTree.open("tree")
        self.assertTrue(tree.is_versioned("b"))
        self.assertTrue(tree.is_versioned("d"))

    def test_mv_auto_one_path(self):
        self.make_abcd_tree()
        out, err = self.run_bzr("mv --auto tree")
        self.assertEqual(out, "")
        self.assertEqual(err, "a => b\nc => d\n")
        tree = workingtree.WorkingTree.open("tree")
        self.assertTrue(tree.is_versioned("b"))
        self.assertTrue(tree.is_versioned("d"))

    def test_mv_auto_two_paths(self):
        self.make_abcd_tree()
        out, err = self.run_bzr("mv --auto tree tree2", retcode=3)
        self.assertEqual("brz: ERROR: Only one path may be specified to --auto.\n", err)

    def test_mv_auto_dry_run(self):
        self.make_abcd_tree()
        out, err = self.run_bzr("mv --auto --dry-run", working_dir="tree")
        self.assertEqual(out, "")
        self.assertEqual(err, "a => b\nc => d\n")
        tree = workingtree.WorkingTree.open("tree")
        self.assertTrue(tree.is_versioned("a"))
        self.assertTrue(tree.is_versioned("c"))

    def test_mv_no_auto_dry_run(self):
        self.make_abcd_tree()
        out, err = self.run_bzr("mv c d --dry-run", working_dir="tree", retcode=3)
        self.assertEqual("brz: ERROR: --dry-run requires --auto.\n", err)

    def test_mv_auto_after(self):
        self.make_abcd_tree()
        out, err = self.run_bzr("mv --auto --after", working_dir="tree", retcode=3)
        self.assertEqual("brz: ERROR: --after cannot be specified with --auto.\n", err)

    def test_mv_quiet(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["aaa"])
        tree.add(["aaa"])
        out, err = self.run_bzr("mv --quiet aaa bbb")
        self.assertEqual(out, "")
        self.assertEqual(err, "")

    def test_mv_readonly_lightweight_checkout(self):
        branch = self.make_branch("foo")
        branch = breezy.branch.Branch.open(self.get_readonly_url("foo"))
        tree = branch.create_checkout("tree", lightweight=True)
        self.build_tree(["tree/path"])
        tree.add("path")
        # If this fails, the tree is trying to acquire a branch lock, which it
        # shouldn't.
        self.run_bzr(["mv", "tree/path", "tree/path2"])

    def test_mv_unversioned_non_ascii(self):
        """Clear error on mv of an unversioned non-ascii file, see lp:707954."""
        self.requireFeature(UnicodeFilenameFeature)
        self.make_branch_and_tree(".")
        self.build_tree(["\xa7"])
        out, err = self.run_bzr_error(
            ["Could not rename", "not versioned"], ["mv", "\xa7", "b"]
        )

    def test_mv_removed_non_ascii(self):
        """Clear error on mv of a removed non-ascii file, see lp:898541."""
        self.requireFeature(UnicodeFilenameFeature)
        tree = self.make_branch_and_tree(".")
        self.build_tree(["\xa7"])
        tree.add(["\xa7"])
        tree.commit("Adding \xa7")
        os.remove("\xa7")
        out, err = self.run_bzr_error(
            ["Could not rename", "not exist"], ["mv", "\xa7", "b"]
        )

    def test_dupe_move(self):
        self.script_runner = script.ScriptRunner()
        self.script_runner.run_script(
            self,
            """
        $ brz init brz-bug
        Created a standalone tree (format: 2a)
        $ cd brz-bug
        $ mkdir dir
        $ brz add
        adding dir
        $ echo text >> dir/test.txt
        $ brz add
        adding dir/test.txt
        $ brz ci -m "Add files"
        2>Committing to: .../brz-bug/
        2>added dir
        2>added dir/test.txt
        2>Committed revision 1.
        $ mv dir dir2
        $ mv dir2/test.txt dir2/test2.txt
        $ brz st
        removed:
          dir/
          dir/test.txt
        unknown:
          dir2/
        $ brz mv dir dir2
        dir => dir2
        $ brz st
        removed:
          dir/test.txt
        renamed:
          dir/ => dir2/
        unknown:
          dir2/test2.txt
        $ brz mv dir/test.txt dir2/test2.txt
        dir/test.txt => dir2/test2.txt
        """,
        )



# Copyright (C) 2006-2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""UI tests for brz ignore."""

import os
import re

from breezy import ignores
from breezy.tests import TestCaseWithTransport


class TestCommands(TestCaseWithTransport):
    def test_ignore_absolutes(self):
        """'ignore' with an absolute path returns an error."""
        self.make_branch_and_tree(".")
        self.run_bzr_error(
            ("brz: ERROR: NAME_PATTERN should not be an absolute path\n",),
            "ignore /crud",
        )

    def test_ignore_directories(self):
        """Ignoring a directory should ignore directory tree.

        Also check that trailing slashes on directories are stripped.
        """
        self.run_bzr("init")
        self.build_tree(["dir1/", "dir1/foo", "dir2/", "dir2/bar", "dir3/", "dir3/baz"])
        self.run_bzr(["ignore", "dir1", "dir2/", "dir4\\"])
        self.check_file_contents(".bzrignore", b"dir1\ndir2\ndir4\n")
        self.assertEqual(self.run_bzr("unknowns")[0], "dir3\n")

    def test_ignore_patterns(self):
        tree = self.make_branch_and_tree(".")

        self.assertEqual(list(tree.unknowns()), [])

        # is_ignored() will now create the user global ignore file
        # if it doesn't exist, so make sure we ignore it in our tests
        ignores._set_user_ignores(["*.tmp"])

        self.build_tree_contents([("foo.tmp", b".tmp files are ignored by default")])
        self.assertEqual(list(tree.unknowns()), [])

        self.build_tree_contents([("foo.c", b"int main() {}")])
        self.assertEqual(list(tree.unknowns()), ["foo.c"])

        tree.add("foo.c")
        self.assertEqual(list(tree.unknowns()), [])

        # 'ignore' works when creating the .bzrignore file
        self.build_tree_contents([("foo.blah", b"blah")])
        self.assertEqual(list(tree.unknowns()), ["foo.blah"])
        self.run_bzr("ignore *.blah")
        self.assertEqual(list(tree.unknowns()), [])
        self.check_file_contents(".bzrignore", b"*.blah\n")

        # 'ignore' works when then .bzrignore file already exists
        self.build_tree_contents([("garh", b"garh")])
        self.assertEqual(list(tree.unknowns()), ["garh"])
        self.run_bzr("ignore garh")
        self.assertEqual(list(tree.unknowns()), [])
        self.check_file_contents(".bzrignore", b"*.blah\ngarh\n")

    def test_ignore_multiple_arguments(self):
        """'ignore' works with multiple arguments."""
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b", "c", "d"])
        self.assertEqual(list(tree.unknowns()), ["a", "b", "c", "d"])
        self.run_bzr("ignore a b c")
        self.assertEqual(list(tree.unknowns()), ["d"])
        self.check_file_contents(".bzrignore", b"a\nb\nc\n")

    def test_ignore_no_arguments(self):
        """'ignore' with no arguments returns an error."""
        self.make_branch_and_tree(".")
        self.run_bzr_error(
            (
                "brz: ERROR: ignore requires at least one "
                "NAME_PATTERN or --default-rules.\n",
            ),
            "ignore",
        )

    def test_ignore_default_rules(self):
        out, err = self.run_bzr(["ignore", "--default-rules"])
        reference_set = set(ignores.USER_DEFAULTS)
        output_set = set(out.rstrip().split("\n"))
        self.assertEqual(reference_set, output_set)
        self.assertEqual("", err)

    def test_ignore_versioned_file(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b"])
        tree.add("a")

        # test a single versioned file
        out, err = self.run_bzr("ignore a")
        self.assertEqual(
            out,
            "Warning: the following files are version controlled"
            " and match your ignore pattern:\na\n"
            "These files will continue to be version controlled"
            " unless you 'brz remove' them.\n",
        )

        # test a single unversioned file
        out, err = self.run_bzr("ignore b")
        self.assertEqual(out, "")

        # test wildcards
        tree.add("b")
        out, err = self.run_bzr("ignore *")
        self.assertEqual(
            out,
            "Warning: the following files are version controlled"
            " and match your ignore pattern:\n.bzrignore\na\nb\n"
            "These files will continue to be version controlled"
            " unless you 'brz remove' them.\n",
        )

    def test_ignored_versioned_file_matching_new_pattern(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["a", "b"])
        tree.add(["a", "b"])
        self.run_bzr("ignore *")

        # If only the given pattern is used then only 'b' should match in
        # this case.
        out, err = self.run_bzr("ignore b")
        self.assertEqual(
            out,
            "Warning: the following files are version controlled"
            " and match your ignore pattern:\nb\n"
            "These files will continue to be version controlled"
            " unless you 'brz remove' them.\n",
        )

    def test_ignore_directory(self):
        """Test --directory option."""
        self.make_branch_and_tree("a")
        self.run_bzr(["ignore", "--directory=a", "README"])
        self.check_file_contents("a/.bzrignore", b"README\n")

    def test_ignored_invalid_pattern(self):
        """Ensure graceful handling for invalid ignore pattern.

        Test case for #300062.
        Invalid pattern should show clear error message.
        Invalid pattern should not be added to .bzrignore file.
        """
        self.make_branch_and_tree(".")
        out, err = self.run_bzr(["ignore", "RE:*.cpp", "foo", "RE:["], 3)
        self.assertEqual(out, "")
        self.assertContainsRe(
            err, r"Invalid ignore pattern.*RE:\*\.cpp.*RE:\[", re.DOTALL
        )
        self.assertNotContainsRe(err, "foo", re.DOTALL)
        self.assertFalse(os.path.isfile(".bzrignore"))



from ...revision import Revision
from ...tests import TestCase, TestCaseWithTransport
from .cmds import collapse_by_person, get_revisions_and_committers


class TestGetRevisionsAndCommitters(TestCaseWithTransport):
    def test_simple(self):
        wt = self.make_branch_and_tree(".")
        wt.commit(message="1", committer="Fero <fero@example.com>", rev_id=b"1")
        wt.commit(message="2", committer="Fero <fero@example.com>", rev_id=b"2")
        wt.commit(message="3", committer="Jano <jano@example.com>", rev_id=b"3")
        wt.commit(
            message="4",
            committer="Jano <jano@example.com>",
            authors=["Vinco <vinco@example.com>"],
            rev_id=b"4",
        )
        wt.commit(message="5", committer="Ferko <fero@example.com>", rev_id=b"5")
        revs, committers = get_revisions_and_committers(
            wt.branch.repository, [b"1", b"2", b"3", b"4", b"5"]
        )
        fero = ("Fero", "fero@example.com")
        jano = ("Jano", "jano@example.com")
        vinco = ("Vinco", "vinco@example.com")
        ferok = ("Ferko", "fero@example.com")
        self.assertEqual(
            {fero: fero, jano: jano, vinco: vinco, ferok: fero}, committers
        )

    def test_empty_email(self):
        wt = self.make_branch_and_tree(".")
        wt.commit(message="1", committer="Fero", rev_id=b"1")
        wt.commit(message="2", committer="Fero", rev_id=b"2")
        wt.commit(message="3", committer="Jano", rev_id=b"3")
        revs, committers = get_revisions_and_committers(
            wt.branch.repository, [b"1", b"2", b"3"]
        )
        self.assertEqual(
            {
                ("Fero", ""): ("Fero", ""),
                ("Jano", ""): ("Jano", ""),
            },
            committers,
        )

    def test_different_case(self):
        wt = self.make_branch_and_tree(".")
        wt.commit(message="1", committer="Fero", rev_id=b"1")
        wt.commit(message="2", committer="Fero", rev_id=b"2")
        wt.commit(message="3", committer="FERO", rev_id=b"3")
        revs, committers = get_revisions_and_committers(
            wt.branch.repository, [b"1", b"2", b"3"]
        )
        self.assertEqual(
            {
                ("Fero", ""): ("Fero", ""),
                ("FERO", ""): ("Fero", ""),
            },
            committers,
        )
        self.assertEqual([b"1", b"2", b"3"], sorted([r.revision_id for r in revs]))


class TestCollapseByPerson(TestCase):
    def test_no_conflicts(self):
        revisions = [
            Revision("1", {}, committer="Foo <foo@example.com>"),
            Revision("2", {}, committer="Bar <bar@example.com>"),
            Revision("3", {}, committer="Bar <bar@example.com>"),
        ]
        foo = ("Foo", "foo@example.com")
        bar = ("Bar", "bar@example.com")
        committers = {foo: foo, bar: bar}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(2, info[0][0])
        self.assertEqual({"bar@example.com": 2}, info[0][2])
        self.assertEqual({"Bar": 2}, info[0][3])

    def test_different_email(self):
        revisions = [
            Revision("1", {}, committer="Foo <foo@example.com>"),
            Revision("2", {}, committer="Foo <bar@example.com>"),
            Revision("3", {}, committer="Foo <bar@example.com>"),
        ]
        foo = ("Foo", "foo@example.com")
        bar = ("Foo", "bar@example.com")
        committers = {foo: foo, bar: foo}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(3, info[0][0])
        self.assertEqual({"foo@example.com": 1, "bar@example.com": 2}, info[0][2])
        self.assertEqual({"Foo": 3}, info[0][3])

    def test_different_name(self):
        revisions = [
            Revision("1", {}, committer="Foo <foo@example.com>"),
            Revision("2", {}, committer="Bar <foo@example.com>"),
            Revision("3", {}, committer="Bar <foo@example.com>"),
        ]
        foo = ("Foo", "foo@example.com")
        bar = ("Bar", "foo@example.com")
        committers = {foo: foo, bar: foo}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(3, info[0][0])
        self.assertEqual({"foo@example.com": 3}, info[0][2])
        self.assertEqual({"Foo": 1, "Bar": 2}, info[0][3])

    def test_different_name_case(self):
        revisions = [
            Revision("1", {}, committer="Foo <foo@example.com>"),
            Revision("2", {}, committer="Foo <foo@example.com>"),
            Revision("3", {}, committer="FOO <bar@example.com>"),
        ]
        foo = ("Foo", "foo@example.com")
        FOO = ("FOO", "bar@example.com")
        committers = {foo: foo, FOO: foo}
        info = collapse_by_person(revisions, committers)
        self.assertEqual(3, info[0][0])
        self.assertEqual({"foo@example.com": 2, "bar@example.com": 1}, info[0][2])
        self.assertEqual({"Foo": 2, "FOO": 1}, info[0][3])



# Copyright (C) 2018 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for GitLab."""

import json
import os
import re
import time
from datetime import datetime
from typing import Optional

from ... import bedding, controldir, errors, urlutils
from ... import branch as _mod_branch
from ...forge import (
    Forge,
    ForgeLoginRequired,
    MergeProposal,
    MergeProposalBuilder,
    MergeProposalExists,
    NoSuchProject,
    PrerequisiteBranchUnsupported,
    SourceNotDerivedFromTarget,
    UnsupportedForge,
    determine_title,
)
from ...git.urls import git_url_to_bzr_url
from ...trace import mutter
from ...transport import get_transport

_DEFAULT_FILES = ["/etc/python-gitlab.cfg", "~/.python-gitlab.cfg"]
DEFAULT_PAGE_SIZE = 50
DEFAULT_PREFERRED_SCHEMES = ["ssh", "http"]
SCHEME_MAP = {
    "git+ssh": "ssh_url_to_repo",
    "ssh": "ssh_url_to_repo",
    "http": "http_url_to_repo",
    "https": "http_url_to_repo",
}


def mp_status_to_status(status):
    return {"all": "all", "open": "opened", "merged": "merged", "closed": "closed"}[
        status
    ]


def parse_timestring(ts):
    return datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%fZ")


class NotGitLabUrl(errors.BzrError):
    _fmt = "Not a GitLab URL: %(url)s"

    def __init__(self, url):
        errors.BzrError.__init__(self)
        self.url = url


class NotMergeRequestUrl(errors.BzrError):
    _fmt = "Not a merge proposal URL: %(url)s"

    def __init__(self, host, url):
        errors.BzrError.__init__(self)
        self.host = host
        self.url = url


class GitLabError(errors.BzrError):
    _fmt = "GitLab error: %(error)s"

    def __init__(self, error, full_response):
        errors.BzrError.__init__(self)
        self.error = error
        self.full_response = full_response


class GitLabUnprocessable(errors.BzrError):
    _fmt = "GitLab can not process request: %(error)s."

    def __init__(self, error, full_response):
        errors.BzrError.__init__(self)
        self.error = error
        self.full_response = full_response


class DifferentGitLabInstances(errors.BzrError):
    _fmt = (
        "Can't create merge proposals across GitLab instances: "
        "%(source_host)s and %(target_host)s"
    )

    def __init__(self, source_host, target_host):
        self.source_host = source_host
        self.target_host = target_host


class GitLabLoginMissing(ForgeLoginRequired):
    _fmt = "Please log into GitLab instance at %(forge)s"


class GitlabLoginError(errors.BzrError):
    _fmt = "Error logging in: %(error)s"

    def __init__(self, error):
        self.error = error


class GitLabConflict(errors.BzrError):
    _fmt = "Conflict during operation: %(reason)s"

    def __init__(self, reason):
        errors.BzrError(self)
        self.reason = reason


class ForkingDisabled(errors.BzrError):
    _fmt = "Forking on project %(project)s is disabled."

    def __init__(self, project):
        self.project = project


class MergeRequestConflict(Exception):
    """Raised when a merge requests conflicts."""

    def __init__(self, reason):
        self.reason = reason


class ProjectCreationTimeout(errors.BzrError):
    _fmt = "Timeout (%(timeout)ds) while waiting for project %(project)s to be created."

    def __init__(self, project, timeout):
        self.project = project
        self.timeout = timeout


def store_gitlab_token(name, url, private_token):
    """Store a GitLab token in a configuration file."""
    from breezy.config import AuthenticationConfig

    auth_config = AuthenticationConfig()
    auth_config._set_option(name, "url", url)
    auth_config._set_option(name, "forge", "gitlab")
    auth_config._set_option(name, "private_token", private_token)


def iter_tokens():
    import configparser

    config = configparser.ConfigParser()
    config.read(
        [os.path.expanduser(p) for p in _DEFAULT_FILES]
        +
        # backwards compatibility
        [os.path.join(bedding.config_dir(), "gitlab.conf")]
    )
    for name, creds in config.items():
        if "url" not in creds:
            continue
        yield name, creds

    from breezy.config import AuthenticationConfig

    auth_config = AuthenticationConfig()
    for name, creds in auth_config._get_config().iteritems():
        if creds.get("forge") == "gitlab":
            yield name, creds
        else:
            url = creds.get("url")
            # Hack for those without forge set
            if url and url.startswith("https://gitlab.com/"):
                yield name, creds


def get_credentials_by_url(url):
    for _name, credentials in iter_tokens():
        if credentials["url"].rstrip("/") == url.rstrip("/"):
            return credentials
    else:
        return None


def parse_gitlab_url(url):
    (scheme, user, password, host, port, path) = urlutils.parse_url(url)
    if scheme not in ("git+ssh", "https", "http"):
        raise NotGitLabUrl(url)
    if not host:
        raise NotGitLabUrl(url)
    path = path.strip("/")
    if path.endswith(".git"):
        path = path[:-4]
    return host, path


def parse_gitlab_branch_url(branch):
    url = urlutils.strip_segment_parameters(branch.user_url)
    host, path = parse_gitlab_url(url)
    return host, path, branch.name


def parse_gitlab_merge_request_url(url):
    (scheme, user, password, host, port, path) = urlutils.parse_url(url)
    if scheme not in ("git+ssh", "https", "http"):
        raise NotGitLabUrl(url)
    if not host:
        raise NotGitLabUrl(url)
    path = path.strip("/")
    parts = path.split("/")
    if len(parts) < 2:
        raise NotMergeRequestUrl(host, url)
    if parts[-2] != "merge_requests":
        raise NotMergeRequestUrl(host, url)
    if parts[-3] == "-":
        project_name = "/".join(parts[:-3])
    else:
        project_name = "/".join(parts[:-2])
    return host, project_name, int(parts[-1])


def _unexpected_status(path, response):
    raise errors.UnexpectedHttpStatus(
        path,
        response.status,
        response.data.decode("utf-8", "replace"),
        headers=response.getheaders(),
    )


class GitLabMergeProposal(MergeProposal):
    supports_auto_merge = True

    def __init__(self, gl, mr):
        self.gl = gl
        self._mr = mr

    def _update(self, **kwargs):
        try:
            self.gl._update_merge_request(
                self._mr["project_id"], self._mr["iid"], kwargs
            )
        except GitLabConflict as e:
            self.gl._handle_merge_request_conflict(
                e.reason, self.get_source_branch_url(), self._mr["target_project_id"]
            )

    def __repr__(self):
        return "<{} at {!r}>".format(type(self).__name__, self._mr["web_url"])

    @property
    def url(self):
        return self._mr["web_url"]

    def get_web_url(self):
        return self._mr["web_url"]

    def get_description(self):
        return self._mr["description"]

    def set_description(self, description):
        try:
            self._update(description=description)
        except errors.UnexpectedHttpStatus as e:
            # HACK: Some versions of GitLab apply the changes but fail with a 500
            # This appears to happen at least with version 15.5.6
            if e.code != 500:
                raise
            self._mr = self.gl._get_merge_request(
                self._mr["project_id"], self._mr["iid"]
            )
            if self._mr["description"] != description:
                raise

    def get_commit_message(self):
        return self._mr.get("merge_commit_message")

    def set_commit_message(self, message):
        raise errors.UnsupportedOperation(self.set_commit_message, self)

    def get_title(self):
        return self._mr.get("title")

    def set_title(self, title):
        self._update(title=title)

    def _branch_url_from_project(
        self, project_id, branch_name, *, preferred_schemes=None
    ):
        if project_id is None:
            return None
        project = self.gl._get_project(project_id)
        if preferred_schemes is None:
            preferred_schemes = DEFAULT_PREFERRED_SCHEMES
        for scheme in preferred_schemes:
            if scheme in SCHEME_MAP:
                return gitlab_url_to_bzr_url(project[SCHEME_MAP[scheme]], branch_name)
        raise KeyError

    def get_source_branch_url(self, *, preferred_schemes=None):
        return self._branch_url_from_project(
            self._mr["source_project_id"],
            self._mr["source_branch"],
            preferred_schemes=preferred_schemes,
        )

    def get_source_revision(self):
        from breezy.git.mapping import default_mapping

        sha = self._mr["sha"]
        if sha is None:
            return None
        return default_mapping.revision_id_foreign_to_bzr(sha.encode("ascii"))

    def get_target_branch_url(self, *, preferred_schemes=None):
        return self._branch_url_from_project(
            self._mr["target_project_id"],
            self._mr["target_branch"],
            preferred_schemes=preferred_schemes,
        )

    def set_target_branch_name(self, name):
        self._update(target_branch=name)

    def _get_project_name(self, project_id):
        source_project = self.gl._get_project(project_id)
        return source_project["path_with_namespace"]

    def get_source_project(self):
        return self._get_project_name(self._mr["source_project_id"])

    def get_target_project(self):
        return self._get_project_name(self._mr["target_project_id"])

    def is_merged(self):
        return self._mr["state"] == "merged"

    def is_closed(self):
        return self._mr["state"] == "closed"

    def reopen(self):
        return self._update(state_event="reopen")

    def close(self):
        self._update(state_event="close")

    def merge(self, commit_message=None, auto=False):
        # https://docs.gitlab.com/ee/api/merge_requests.html#accept-mr
        ret = self.gl._merge_mr(
            self._mr["project_id"],
            self._mr["iid"],
            kwargs={
                "merge_commit_message": commit_message,
                "merge_when_pipeline_succeeds": auto,
            },
        )
        self._mr.update(ret)

    def can_be_merged(self):
        if self._mr["merge_status"] == "cannot_be_merged":
            return False
        elif self._mr["merge_status"] == "can_be_merged":
            return True
        elif self._mr["merge_status"] in (
            "unchecked",
            "cannot_be_merged_recheck",
            "checking",
        ):
            # See https://gitlab.com/gitlab-org/gitlab/-/commit/7517105303c for
            # an explanation of the distinction between unchecked and
            # cannot_be_merged_recheck
            return None
        else:
            raise ValueError(self._mr["merge_status"])

    def get_merged_by(self):
        user = self._mr.get("merge_user")
        if user is None:
            return None
        return user["username"]

    def get_merged_at(self):
        merged_at = self._mr.get("merged_at")
        if merged_at is None:
            return None
        return parse_timestring(merged_at)

    def post_comment(self, body):
        kwargs = {"body": body}
        self.gl._post_merge_request_note(
            self._mr["project_id"], self._mr["iid"], kwargs
        )


def gitlab_url_to_bzr_url(url, name):
    return git_url_to_bzr_url(url, branch=name)


class GitLab(Forge):
    """GitLab forge implementation."""

    supports_merge_proposal_labels = True
    supports_merge_proposal_title = True
    supports_merge_proposal_commit_message = False
    supports_allow_collaboration = True
    merge_proposal_description_format = "markdown"

    def __repr__(self):
        return "<GitLab({!r})>".format(self.base_url)

    @property
    def base_url(self):
        return self.transport.base

    @property
    def base_hostname(self):
        return urlutils.parse_url(self.base_url)[3]

    def _find_correct_project_name(self, path):
        try:
            resp = self.transport.request(
                "GET", urlutils.join(self.base_url, path), headers=self.headers
            )
        except errors.RedirectRequested as e:
            return urlutils.parse_url(e.target)[-1].strip("/")
        if resp.status != 200:
            _unexpected_status(path, resp)
        return None

    def _api_request(self, method, path, fields=None, body=None):
        return self.transport.request(
            method,
            urlutils.join(self.base_url, "api", "v4", path),
            headers=self.headers,
            fields=fields,
            body=body,
        )

    def __init__(self, transport, private_token):
        self.transport = transport
        self.headers = {"Private-Token": private_token}
        self._current_user = None

    def _get_user(self, username):
        path = "users/{}".format(urlutils.quote(str(username), ""))
        response = self._api_request("GET", path)
        if response.status == 404:
            raise KeyError("no such user {}".format(username))
        if response.status == 200:
            return json.loads(response.data)
        _unexpected_status(path, response)

    def _get_user_by_email(self, email):
        path = "users?search={}".format(urlutils.quote(str(email), ""))
        response = self._api_request("GET", path)
        if response.status == 404:
            raise KeyError("no such user {}".format(email))
        if response.status == 200:
            ret = json.loads(response.data)
            if len(ret) != 1:
                raise ValueError("unexpected number of results; {!r}".format(ret))
            return ret[0]
        _unexpected_status(path, response)

    def _get_project(self, project_name, _redirect_checked=False):
        path = "projects/{}".format(urlutils.quote(str(project_name), ""))
        response = self._api_request("GET", path)
        if response.status == 404:
            if not _redirect_checked:
                project_name = self._find_correct_project_name(project_name)
                if project_name is not None:
                    return self._get_project(project_name, _redirect_checked=True)
            raise NoSuchProject(project_name)
        if response.status == 200:
            return json.loads(response.data)
        _unexpected_status(path, response)

    def _get_namespace(self, namespace):
        path = "namespaces/" + urlutils.quote(str(namespace), "")
        response = self._api_request("GET", path)
        if response.status == 200:
            return json.loads(response.data)
        if response.status == 404:
            return None
        _unexpected_status(path, response)

    def create_project(self, project_name, summary=None):
        if project_name.endswith(".git"):
            project_name = project_name[:-4]
        if "/" in project_name:
            namespace_path, path = project_name.lstrip("/").rsplit("/", 1)
        else:
            namespace_path = ""
            path = project_name

        namespace = self._get_namespace(namespace_path)
        if namespace is None:
            raise Exception("namespace {} does not exist".format(namespace_path))

        fields = {
            "path": path,
            "namespace_id": namespace["id"],
        }
        if summary is not None:
            fields["description"] = summary
        response = self._api_request("POST", "projects", fields=fields)
        if response.status == 400:
            ret = json.loads(response.data)
            if ret.get("message", {}).get("path") == ["has already been taken"]:
                raise errors.AlreadyControlDirError(project_name)
            raise
        if response.status == 403:
            raise errors.PermissionDenied(response.text)
        if response.status not in (200, 201):
            _unexpected_status("projects", response)
        project = json.loads(response.data)
        return project

    def fork_project(self, project_name, timeout=50, interval=5, owner=None):
        path = "projects/{}/fork".format(urlutils.quote(str(project_name), ""))
        fields = {}
        if owner is not None:
            fields["namespace"] = owner
        response = self._api_request("POST", path, fields=fields)
        if response.status == 404:
            raise ForkingDisabled(project_name)
        if response.status == 409:
            resp = json.loads(response.data)
            raise GitLabConflict(resp.get("message"))
        if response.status not in (200, 201):
            _unexpected_status(path, response)
        # The response should be valid JSON, but let's ignore it
        project = json.loads(response.data)
        # Spin and wait until import_status for new project
        # is complete.
        deadline = time.time() + timeout
        while project["import_status"] not in ("finished", "none"):
            mutter("import status is %s", project["import_status"])
            if time.time() > deadline:
                raise ProjectCreationTimeout(project["path_with_namespace"], timeout)
            time.sleep(interval)
            project = self._get_project(project["path_with_namespace"])
        return project

    def _handle_merge_request_conflict(self, message, source_url, target_project):
        m = re.fullmatch(
            r"Another open merge request already exists for "
            r"this source branch: \!([0-9]+)",
            message[0],
        )
        if m:
            merge_id = int(m.group(1))
            mr = self._get_merge_request(target_project, merge_id)
            raise MergeProposalExists(source_url, GitLabMergeProposal(self, mr))
        raise MergeRequestConflict(message)

    def get_current_user(self):
        if not self._current_user:
            self._retrieve_user()
        return self._current_user["username"]

    def get_user_url(self, username):
        return urlutils.join(self.base_url, username)

    def _list_paged(self, path, parameters=None, per_page=None):
        if parameters is None:
            parameters = {}
        else:
            parameters = dict(parameters.items())
        if per_page:
            parameters["per_page"] = str(per_page)
        page = "1"
        while page:
            parameters["page"] = page
            response = self._api_request(
                "GET",
                path
                + "?"
                + "&".join(["{}={}".format(*item) for item in parameters.items()]),
            )
            if response.status == 403:
                raise errors.PermissionDenied(response.text)
            if response.status != 200:
                _unexpected_status(path, response)
            page = response.getheader("X-Next-Page")
            yield from json.loads(response.data)

    def _list_merge_requests(self, author=None, project=None, state=None):
        if project is not None:
            path = "projects/{}/merge_requests".format(urlutils.quote(str(project), ""))
        else:
            path = "merge_requests"
        parameters = {}
        if state:
            parameters["state"] = state
        if author:
            parameters["author_username"] = urlutils.quote(author, "")
        return self._list_paged(path, parameters, per_page=DEFAULT_PAGE_SIZE)

    def _get_merge_request(self, project, merge_id):
        path = "projects/%s/merge_requests/%d" % (
            urlutils.quote(str(project), ""),
            merge_id,
        )
        response = self._api_request("GET", path)
        if response.status == 403:
            raise errors.PermissionDenied(response.text)
        if response.status != 200:
            _unexpected_status(path, response)
        return json.loads(response.data)

    def _list_projects(self, owner):
        path = "users/{}/projects".format(urlutils.quote(str(owner), ""))
        parameters = {}
        return self._list_paged(path, parameters, per_page=DEFAULT_PAGE_SIZE)

    def _update_merge_request(self, project_id, iid, mr):
        path = "projects/{}/merge_requests/{}".format(
            urlutils.quote(str(project_id), ""), iid
        )
        response = self._api_request("PUT", path, fields=mr)
        if response.status == 200:
            return json.loads(response.data)
        if response.status == 409:
            raise GitLabConflict(json.loads(response.data).get("message"))
        if response.status == 403:
            raise errors.PermissionDenied(response.text)
        _unexpected_status(path, response)

    def _merge_mr(self, project_id, iid, kwargs):
        path = "projects/{}/merge_requests/{}/merge".format(
            urlutils.quote(str(project_id), ""), iid
        )
        response = self._api_request("PUT", path, fields=kwargs)
        if response.status == 200:
            return json.loads(response.data)
        if response.status == 403:
            raise errors.PermissionDenied(response.text)
        _unexpected_status(path, response)

    def _post_merge_request_note(self, project_id, iid, kwargs):
        path = "projects/{}/merge_requests/{}/notes".format(
            urlutils.quote(str(project_id), ""), iid
        )
        response = self._api_request("POST", path, fields=kwargs)
        if response.status == 201:
            json.loads(response.data)
            return
        if response.status == 403:
            raise errors.PermissionDenied(response.text)
        _unexpected_status(path, response)

    def _create_mergerequest(
        self,
        title,
        source_project_id,
        target_project_id,
        source_branch_name,
        target_branch_name,
        description,
        labels=None,
        allow_collaboration=False,
    ):
        path = "projects/{}/merge_requests".format(source_project_id)
        fields = {
            "title": title,
            "source_branch": source_branch_name,
            "target_branch": target_branch_name,
            "target_project_id": target_project_id,
            "description": description,
            "allow_collaboration": allow_collaboration,
        }
        if labels:
            fields["labels"] = labels
        response = self._api_request("POST", path, fields=fields)
        if response.status == 400:
            data = json.loads(response.data)
            raise GitLabError(data.get("message"), response)
        if response.status == 403:
            raise errors.PermissionDenied(response.text)
        if response.status == 409:
            raise GitLabConflict(json.loads(response.data).get("message"))
        if response.status == 422:
            data = json.loads(response.data)
            raise GitLabUnprocessable(data.get("error") or data.get("message"), data)
        if response.status != 201:
            _unexpected_status(path, response)
        return json.loads(response.data)

    def get_push_url(self, branch):
        (host, project_name, branch_name) = parse_gitlab_branch_url(branch)
        project = self._get_project(project_name)
        return gitlab_url_to_bzr_url(project["ssh_url_to_repo"], branch_name)

    def get_web_url(self, branch):
        (host, project_name, branch_name) = parse_gitlab_branch_url(branch)
        project = self._get_project(project_name)
        if branch_name:
            # TODO(jelmer): Use API to get this URL
            return project["web_url"] + "/-/tree/" + branch_name
        else:
            return project["web_url"]

    def publish_derived(
        self,
        local_branch,
        base_branch,
        name,
        project=None,
        owner=None,
        revision_id=None,
        overwrite=False,
        allow_lossy=True,
        tag_selector=None,
    ):
        if tag_selector is None:

            def tag_selector(t):
                return False

        (host, base_project_name, base_branch_name) = parse_gitlab_branch_url(
            base_branch
        )
        if owner is None:
            owner = base_branch.get_config_stack().get("fork-namespace")
        if owner is None:
            owner = self.get_current_user()
        base_project = self._get_project(base_project_name)
        if project is None:
            project = base_project["path"]
        try:
            target_project = self._get_project("{}/{}".format(owner, project))
        except NoSuchProject:
            target_project = self.fork_project(
                base_project["path_with_namespace"], owner=owner
            )
        remote_repo_url = git_url_to_bzr_url(target_project["ssh_url_to_repo"])
        remote_dir = controldir.ControlDir.open(remote_repo_url)
        try:
            push_result = remote_dir.push_branch(
                local_branch,
                revision_id=revision_id,
                overwrite=overwrite,
                name=name,
                tag_selector=tag_selector,
            )
        except errors.NoRoundtrippingSupport:
            if not allow_lossy:
                raise
            push_result = remote_dir.push_branch(
                local_branch,
                revision_id=revision_id,
                overwrite=overwrite,
                name=name,
                lossy=True,
                tag_selector=tag_selector,
            )
        public_url = gitlab_url_to_bzr_url(target_project["http_url_to_repo"], name)
        return push_result.target_branch, public_url

    def get_derived_branch(
        self, base_branch, name, project=None, owner=None, preferred_schemes=None
    ):
        (host, base_project, base_branch_name) = parse_gitlab_branch_url(base_branch)
        if owner is None:
            owner = self.get_current_user()
        if project is None:
            project = self._get_project(base_project)["path"]
        try:
            target_project = self._get_project("{}/{}".format(owner, project))
        except NoSuchProject:
            raise errors.NotBranchError(
                "{}/{}/{}".format(self.base_url, owner, project)
            )
        if preferred_schemes is None:
            preferred_schemes = ["git+ssh"]
        for scheme in preferred_schemes:
            if scheme == "git+ssh":
                gitlab_url = target_project["ssh_url_to_repo"]
                break
            elif scheme == "https":
                gitlab_url = target_project["http_url_to_repo"]
                break
        else:
            raise AssertionError
        return _mod_branch.Branch.open(
            gitlab_url_to_bzr_url(gitlab_url, name),
            possible_transports=[base_branch.user_transport],
        )

    def get_proposer(self, source_branch, target_branch):
        return GitlabMergeProposalBuilder(self, source_branch, target_branch)

    def iter_proposals(self, source_branch, target_branch, status):
        (source_host, source_project_name, source_branch_name) = (
            parse_gitlab_branch_url(source_branch)
        )
        (target_host, target_project_name, target_branch_name) = (
            parse_gitlab_branch_url(target_branch)
        )
        if source_host != target_host:
            raise DifferentGitLabInstances(source_host, target_host)
        source_project = self._get_project(source_project_name)
        target_project = self._get_project(target_project_name)
        state = mp_status_to_status(status)
        for mr in self._list_merge_requests(project=target_project["id"], state=state):
            if (
                mr["source_project_id"] != source_project["id"]
                or mr["source_branch"] != source_branch_name
                or mr["target_project_id"] != target_project["id"]
                or mr["target_branch"] != target_branch_name
            ):
                continue
            yield GitLabMergeProposal(self, mr)

    def hosts(self, branch):
        try:
            (host, project, branch_name) = parse_gitlab_branch_url(branch)
        except NotGitLabUrl:
            return False
        return self.base_hostname == host

    def _retrieve_user(self):
        if self._current_user:
            return
        try:
            response = self._api_request("GET", "user")
        except errors.UnexpectedHttpStatus as e:
            if e.code == 401:
                raise GitLabLoginMissing(self.base_url)
            raise
        if response.status == 200:
            self._current_user = json.loads(response.data)
            return
        if response.status == 401:
            if json.loads(response.data) == {"message": "401 Unauthorized"}:
                raise GitLabLoginMissing(self.base_url)
            else:
                raise GitlabLoginError(response.text)
        raise UnsupportedForge(self.base_url)

    @classmethod
    def probe_from_hostname(cls, hostname, possible_transports=None):
        base_url = "https://{}".format(hostname)
        credentials = get_credentials_by_url(base_url)
        if credentials is not None:
            transport = get_transport(base_url, possible_transports=possible_transports)
            instance = cls(transport, credentials.get("private_token"))
            instance._retrieve_user()
            return instance
        # We could potentially probe for e.g. /api/v4/metadata here
        # But none of the non-project APIs appear to be accessible without
        # authentication :-(
        raise UnsupportedForge(hostname)

    @classmethod
    def probe_from_url(cls, url, possible_transports=None):
        try:
            (host, project) = parse_gitlab_url(url)
        except NotGitLabUrl:
            raise UnsupportedForge(url)
        transport = get_transport(
            "https://{}".format(host), possible_transports=possible_transports
        )
        credentials = get_credentials_by_url(transport.base)
        if credentials is not None:
            instance = cls(transport, credentials.get("private_token"))
            instance._retrieve_user()
            return instance
        try:
            resp = transport.request(
                "GET",
                "https://{}/api/v4/projects/{}".format(
                    host, urlutils.quote(str(project), "")
                ),
            )
        except errors.UnexpectedHttpStatus:
            raise UnsupportedForge(url)
        except errors.RedirectRequested:
            # GitLab doesn't send redirects for these URLs
            raise UnsupportedForge(url)
        else:
            if not resp.getheader("X-Gitlab-Feature-Category"):
                raise UnsupportedForge(url)
            if resp.status in (200, 401):
                raise GitLabLoginMissing("https://{}/".format(host))
            raise UnsupportedForge(url)

    @classmethod
    def iter_instances(cls):
        for _name, credentials in iter_tokens():
            yield cls(
                get_transport(credentials["url"]),
                private_token=credentials.get("private_token"),
            )

    def iter_my_proposals(self, status="open", author=None):
        if author is None:
            author = self.get_current_user()
        state = mp_status_to_status(status)
        for mp in self._list_merge_requests(author=author, state=state):
            yield GitLabMergeProposal(self, mp)

    def iter_my_forks(self, owner: Optional[str] = None):
        if owner is None:
            owner = self.get_current_user()
        for project in self._list_projects(owner=owner):
            base_project = project.get("forked_from_project")
            if not base_project:
                continue
            yield project["path_with_namespace"]

    def get_proposal_by_url(self, url: str) -> GitLabMergeProposal:
        try:
            (host, project, merge_id) = parse_gitlab_merge_request_url(url)
        except NotGitLabUrl:
            raise UnsupportedForge(url)
        except NotMergeRequestUrl as e:
            if self.base_hostname == e.host:
                raise
            else:
                raise UnsupportedForge(url)
        if self.base_hostname != host:
            raise UnsupportedForge(url)
        project = self._get_project(project)
        mr = self._get_merge_request(project["path_with_namespace"], merge_id)
        return GitLabMergeProposal(self, mr)

    def delete_project(self, project):
        path = "projects/{}".format(urlutils.quote(str(project), ""))
        response = self._api_request("DELETE", path)
        if response.status == 404:
            raise NoSuchProject(project)
        if response.status != 202:
            _unexpected_status(path, response)


class GitlabMergeProposalBuilder(MergeProposalBuilder):
    def __init__(self, gl, source_branch, target_branch):
        self.gl = gl
        self.source_branch = source_branch
        (self.source_host, self.source_project_name, self.source_branch_name) = (
            parse_gitlab_branch_url(source_branch)
        )
        self.target_branch = target_branch
        (self.target_host, self.target_project_name, self.target_branch_name) = (
            parse_gitlab_branch_url(target_branch)
        )
        if self.source_host != self.target_host:
            raise DifferentGitLabInstances(self.source_host, self.target_host)

    def get_infotext(self):
        """Determine the initial comment for the merge proposal."""
        info = []
        info.append("Gitlab instance: {}\n".format(self.target_host))
        info.append("Source: {}\n".format(self.source_branch.user_url))
        info.append("Target: {}\n".format(self.target_branch.user_url))
        return "".join(info)

    def get_initial_body(self):
        """Get a body for the proposal for the user to modify.

        :return: a str or None.
        """
        return None

    def create_proposal(
        self,
        description,
        title=None,
        reviewers=None,
        labels=None,
        prerequisite_branch=None,
        commit_message=None,
        work_in_progress=False,
        allow_collaboration=False,
        delete_source_after_merge: Optional[bool] = None,
    ):
        """Perform the submission."""
        # https://docs.gitlab.com/ee/api/merge_requests.html#create-mr
        if prerequisite_branch is not None:
            raise PrerequisiteBranchUnsupported(self)
        # Note that commit_message is ignored, since Gitlab doesn't support it.
        source_project = self.gl._get_project(self.source_project_name)
        target_project = self.gl._get_project(self.target_project_name)
        if title is None:
            title = determine_title(description)
        if work_in_progress:
            title = "WIP: {}".format(title)
        # TODO(jelmer): Allow setting milestone field
        # TODO(jelmer): Allow setting squash field
        kwargs = {
            "title": title,
            "source_project_id": source_project["id"],
            "target_project_id": target_project["id"],
            "source_branch_name": self.source_branch_name,
            "target_branch_name": self.target_branch_name,
            "description": description,
            "allow_collaboration": allow_collaboration,
        }
        if delete_source_after_merge is not None:
            kwargs["should_remove_source_branch"] = delete_source_after_merge
        if labels:
            kwargs["labels"] = ",".join(labels)
        if reviewers:
            kwargs["assignee_ids"] = []
            for reviewer in reviewers:
                if "@" in reviewer:
                    user = self.gl._get_user_by_email(reviewer)
                else:
                    user = self.gl._get_user(reviewer)
                kwargs["assignee_ids"].append(user["id"])
        try:
            merge_request = self.gl._create_mergerequest(**kwargs)
        except GitLabConflict as e:
            self.gl._handle_merge_request_conflict(
                e.reason,
                self.source_branch.user_url,
                target_project["path_with_namespace"],
            )
        except GitLabUnprocessable as e:
            if e.error == ["Source project is not a fork of the target project"]:
                raise SourceNotDerivedFromTarget(self.source_branch, self.target_branch)
            raise
        return GitLabMergeProposal(self.gl, merge_request)


def register_gitlab_instance(shortname, url):
    """Register a gitlab instance.

    :param shortname: Short name (e.g. "gitlab")
    :param url: URL to the gitlab instance
    """
    from breezy.bugtracker import ProjectIntegerBugTracker, tracker_registry

    tracker_registry.register(
        shortname, ProjectIntegerBugTracker(shortname, url + "/{project}/issues/{id}")
    )



# Copyright (C) 2007-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for interface conformance of 'WorkingTree.remove'."""

from breezy import ignores, osutils
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestRemove(TestCaseWithWorkingTree):
    """Tests WorkingTree.remove."""

    files = ["a", "b/", "b/c", "d/"]
    rfiles = ["b/c", "b", "a", "d"]
    backup_files = ["a.~1~", "b.~1~/", "b.~1~/c.~1~", "d.~1~/"]
    backup_files_no_version_dirs = ["a.~1~", "b.~1~/", "b.~1~/c.~1~"]

    def get_tree(self, files):
        tree = self.make_branch_and_tree(".")
        self.build_tree(files)
        self.assertPathExists(files)
        return tree

    def get_committed_tree(self, files, message="Committing"):
        tree = self.get_tree(files)
        tree.add(files)
        tree.commit(message)
        if not tree.has_versioned_directories():
            self.assertInWorkingTree([f for f in files if not f.endswith("/")])
            self.assertPathExists(files)
        else:
            self.assertInWorkingTree(files)
        return tree

    def assertRemovedAndDeleted(self, files):
        self.assertNotInWorkingTree(files)
        self.assertPathDoesNotExist(files)

    def assertRemovedAndNotDeleted(self, files):
        self.assertNotInWorkingTree(files)
        self.assertPathExists(files)

    def test_remove_keep(self):
        """Check that files and directories are unversioned but not deleted."""
        tree = self.get_tree(TestRemove.files)
        tree.add(TestRemove.files)

        tree.remove(TestRemove.files)
        self.assertRemovedAndNotDeleted(TestRemove.files)

    def test_remove_keep_subtree(self):
        """Check that a directory is unversioned but not deleted."""
        tree = self.make_branch_and_tree(".")
        subtree = self.make_branch_and_tree("subtree")
        subtree.commit("")
        tree.add("subtree")

        tree.remove("subtree")
        self.assertRemovedAndNotDeleted("subtree")

    def test_remove_unchanged_files(self):
        """Check that unchanged files are removed and deleted."""
        tree = self.get_committed_tree(TestRemove.files)
        tree.remove(TestRemove.files, keep_files=False)
        self.assertRemovedAndDeleted(TestRemove.files)
        tree._validate()

    def test_remove_added_files(self):
        """Removal of newly added files must back them up."""
        tree = self.get_tree(TestRemove.files)
        tree.add(TestRemove.files)
        tree.remove(TestRemove.files, keep_files=False)
        self.assertNotInWorkingTree(TestRemove.files)
        if tree.has_versioned_directories():
            self.assertPathExists(TestRemove.backup_files)
        else:
            self.assertPathExists(TestRemove.backup_files_no_version_dirs)
        tree._validate()

    def test_remove_changed_file(self):
        """Removal of changed files must back it up."""
        tree = self.get_committed_tree(["a"])
        self.build_tree_contents([("a", b"some other new content!")])
        self.assertInWorkingTree("a")
        tree.remove("a", keep_files=False)
        self.assertNotInWorkingTree(TestRemove.files)
        self.assertPathExists("a.~1~")
        tree._validate()

    def test_remove_deleted_files(self):
        """Check that files are removed if they don't exist any more."""
        tree = self.get_committed_tree(TestRemove.files)
        for f in TestRemove.rfiles:
            osutils.delete_any(f)
        self.assertPathDoesNotExist(TestRemove.files)
        tree.remove(TestRemove.files, keep_files=False)
        self.assertRemovedAndDeleted(TestRemove.files)
        tree._validate()

    def test_remove_renamed_files(self):
        """Check that files are removed even if they are renamed."""
        tree = self.get_committed_tree(TestRemove.files)

        for f in TestRemove.rfiles:
            tree.rename_one(f, f + "x")
        rfilesx = ["bx/cx", "bx", "ax", "dx"]
        self.assertPathExists(rfilesx)

        tree.remove(rfilesx, keep_files=False)
        self.assertRemovedAndDeleted(rfilesx)
        tree._validate()

    def test_remove_renamed_changed_files(self):
        """Check that files that are renamed and changed are backed up."""
        tree = self.get_committed_tree(TestRemove.files)

        for f in TestRemove.rfiles:
            tree.rename_one(f, f + "x")
        rfilesx = ["bx/cx", "bx", "ax", "dx"]
        self.build_tree_contents(
            [("ax", b"changed and renamed!"), ("bx/cx", b"changed and renamed!")]
        )
        self.assertPathExists(rfilesx)

        tree.remove(rfilesx, keep_files=False)
        self.assertNotInWorkingTree(rfilesx)
        self.assertPathExists(["bx.~1~/cx.~1~", "bx.~1~", "ax.~1~"])
        if tree.supports_rename_tracking() or not tree.has_versioned_directories():
            self.assertPathDoesNotExist("dx.~1~")  # unchanged file
        else:
            self.assertPathExists("dx.~1~")  # renamed, so appears changed
        tree._validate()

    def test_force_remove_changed_files(self):
        """Check that changed files are removed and deleted when forced."""
        tree = self.get_tree(TestRemove.files)
        tree.add(TestRemove.files)

        tree.remove(TestRemove.files, keep_files=False, force=True)
        self.assertRemovedAndDeleted(TestRemove.files)
        self.assertPathDoesNotExist(["a.~1~", "b.~1~/", "b.~1~/c", "d.~1~/"])
        tree._validate()

    def test_remove_unknown_files(self):
        """Unknown files shuld be backed up."""
        tree = self.get_tree(TestRemove.files)
        tree.remove(TestRemove.files, keep_files=False)
        self.assertRemovedAndDeleted(TestRemove.files)
        if tree.has_versioned_directories():
            self.assertPathExists(TestRemove.backup_files)
        else:
            self.assertPathExists(TestRemove.backup_files_no_version_dirs)
        tree._validate()

    def test_remove_nonexisting_files(self):
        """Try to delete non-existing files."""
        tree = self.get_tree(TestRemove.files)
        tree.remove([""], keep_files=False)
        tree.remove(["xyz", "abc/def"], keep_files=False)
        tree._validate()

    def test_remove_unchanged_directory(self):
        """Unchanged directories should be deleted."""
        files = ["b/", "b/c", "b/sub_directory/", "b/sub_directory/with_file"]
        tree = self.get_committed_tree(files)
        tree.remove("b", keep_files=False)
        self.assertRemovedAndDeleted("b")
        tree._validate()

    def test_remove_absent_directory(self):
        """Removing a absent directory succeeds without corruption (#150438)."""
        paths = ["a/", "a/b"]
        tree = self.get_committed_tree(paths)
        tree.controldir.root_transport.delete_tree("a")
        tree.remove(["a"])
        self.assertRemovedAndDeleted("b")
        tree._validate()

    def test_remove_unknown_ignored_files(self):
        """Unknown ignored files should be deleted."""
        tree = self.get_committed_tree(["b/"])
        ignores.add_runtime_ignores(["*ignored*"])

        self.build_tree(["unknown_ignored_file"])
        self.assertNotEqual(None, tree.is_ignored("unknown_ignored_file"))
        tree.remove("unknown_ignored_file", keep_files=False)
        self.assertRemovedAndDeleted("unknown_ignored_file")

        self.build_tree(["b/unknown_ignored_file", "b/unknown_ignored_dir/"])
        self.assertNotEqual(None, tree.is_ignored("b/unknown_ignored_file"))
        self.assertNotEqual(None, tree.is_ignored("b/unknown_ignored_dir"))
        tree.remove("b", keep_files=False)
        self.assertRemovedAndDeleted("b")
        tree._validate()

    def test_remove_changed_ignored_files(self):
        """Changed ignored files should be backed up."""
        files = ["an_ignored_file"]
        tree = self.get_tree(files)
        tree.add(files)
        ignores.add_runtime_ignores(["*ignored*"])
        self.assertNotEqual(None, tree.is_ignored(files[0]))

        tree.remove(files, keep_files=False)
        self.assertNotInWorkingTree(files)
        self.assertPathExists("an_ignored_file.~1~")
        tree._validate()

    def test_dont_remove_directory_with_unknowns(self):
        """Directories with unknowns should be backed up."""
        directories = ["a/", "b/", "c/", "c/c/", "c/blah"]
        tree = self.get_committed_tree(directories)

        self.build_tree(["a/unknown_file"])
        tree.remove("a", keep_files=False)
        self.assertPathExists("a.~1~/unknown_file")

        self.build_tree(["b/unknown_directory"])
        tree.remove("b", keep_files=False)
        self.assertPathExists("b.~1~/unknown_directory")

        self.build_tree(["c/c/unknown_file"])
        tree.remove("c/c", keep_files=False)
        self.assertPathExists("c/c.~1~/unknown_file")

        tree.remove("c", keep_files=False)
        self.assertPathExists("c.~1~/")

        self.assertNotInWorkingTree(directories)
        tree._validate()

    def test_force_remove_directory_with_unknowns(self):
        """Unchanged non-empty directories should be deleted when forced."""
        files = ["b/", "b/c"]
        tree = self.get_committed_tree(files)

        other_files = [
            "b/unknown_file",
            "b/sub_directory/",
            "b/sub_directory/with_file",
            "b/sub_directory/sub_directory/",
        ]
        self.build_tree(other_files)

        self.assertInWorkingTree(files)
        self.assertPathExists(files)

        tree.remove("b", keep_files=False, force=True)

        self.assertRemovedAndDeleted(files)
        self.assertRemovedAndDeleted(other_files)
        tree._validate()

    def test_remove_directory_with_changed_file(self):
        """Backup directories with changed files."""
        files = ["b/", "b/c"]
        tree = self.get_committed_tree(files)
        self.build_tree_contents([("b/c", b"some other new content!")])

        tree.remove("b", keep_files=False)
        self.assertPathExists("b.~1~/c.~1~")
        self.assertNotInWorkingTree(files)

    def test_remove_force_directory_with_changed_file(self):
        """Delete directories with changed files when forced."""
        files = ["b/", "b/c"]
        tree = self.get_committed_tree(files)
        self.build_tree_contents([("b/c", b"some other new content!")])

        # see if we can force it now..
        tree.remove("b", keep_files=False, force=True)
        self.assertRemovedAndDeleted(files)
        tree._validate()

    def test_remove_directory_with_changed_emigrated_file(self):
        # As per bug #129880
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([("somedir/",), (b"somedir/file", b"contents")])
        tree.add(["somedir", "somedir/file"])
        tree.commit(message="first")
        self.build_tree_contents([("somedir/file", b"changed")])
        tree.rename_one("somedir/file", "moved-file")
        tree.remove("somedir", keep_files=False)
        self.assertNotInWorkingTree("somedir")
        self.assertPathDoesNotExist("somedir")
        self.assertInWorkingTree("moved-file")
        self.assertPathExists("moved-file")

    def test_remove_directory_with_renames(self):
        """Delete directory with renames in or out."""
        files = ["a/", "a/file", "a/directory/", "a/directory/stuff", "b/"]
        files_to_move = ["a/file", "a/directory/"]

        tree = self.get_committed_tree(files)
        # move stuff from a=>b
        tree.move(["a/file", "a/directory"], to_dir="b")

        moved_files = ["b/file", "b/directory/"]
        self.assertRemovedAndDeleted(files_to_move)
        self.assertInWorkingTree(moved_files)
        self.assertPathExists(moved_files)

        # check if it works with renames out
        tree.remove("a", keep_files=False)
        self.assertRemovedAndDeleted(["a/"])

        # check if it works with renames in
        tree.remove("b", keep_files=False)
        self.assertRemovedAndDeleted(["b/"])
        tree._validate()

    def test_non_cwd(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/dir/", "tree/dir/file"])
        tree.add(["dir", "dir/file"])
        tree.commit("add file")
        tree.remove("dir/", keep_files=False)
        self.assertPathDoesNotExist("tree/dir/file")
        self.assertNotInWorkingTree("tree/dir/file", "tree")
        tree._validate()

    def test_remove_uncommitted_removed_file(self):
        # As per bug #152811
        tree = self.get_committed_tree(["a"])
        tree.remove("a", keep_files=False)
        tree.remove("a", keep_files=False)
        self.assertPathDoesNotExist("a")
        tree._validate()

    def test_remove_file_and_containing_dir(self):
        tree = self.get_committed_tree(["config/", "config/file"])
        tree.remove("config/file", keep_files=False)
        tree.remove("config", keep_files=False)
        self.assertPathDoesNotExist("config/file")
        self.assertPathDoesNotExist("config")
        tree._validate()

    def test_remove_dir_before_bzr(self):
        # As per bug #272648. Note that a file must be present in the directory
        # or the bug doesn't manifest itself.
        tree = self.get_committed_tree([".aaa/", ".aaa/file"])
        tree.remove(".aaa/", keep_files=False)
        self.assertPathDoesNotExist(".aaa/file")
        self.assertPathDoesNotExist(".aaa")
        tree._validate()



# Copyright (C) 2005, 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for different inventory implementations."""

# NOTE: Don't import Inventory here, to make sure that we don't accidentally
# hardcode that when we should be using self.make_inventory

from breezy import errors, osutils
from breezy.bzr import inventory
from breezy.bzr.inventory import (
    InventoryFile,
    InventoryLink,
)
from breezy.bzr.tests.per_inventory import TestCaseWithInventory


class TestInventory(TestCaseWithInventory):
    def make_init_inventory(self):
        inv = inventory.Inventory(b"tree-root")
        inv.revision = b"initial-rev"
        inv.root.revision = b"initial-rev"
        return self.inv_to_test_inv(inv)

    def make_file(
        self, file_id, name, parent_id, content=b"content\n", revision=b"new-test-rev"
    ):
        ie = InventoryFile(file_id, name, parent_id)
        ie.text_sha1 = osutils.sha_string(content)
        ie.text_size = len(content)
        ie.revision = revision
        return ie

    def make_link(self, file_id, name, parent_id, target="link-target\n"):
        ie = InventoryLink(file_id, name, parent_id)
        ie.symlink_target = target
        return ie

    def prepare_inv_with_nested_dirs(self):
        inv = inventory.Inventory(b"tree-root")
        inv.root.revision = b"revision"
        for args in [
            ("src", "directory", b"src-id"),
            ("doc", "directory", b"doc-id"),
            ("src/hello.c", "file", b"hello-id"),
            ("src/bye.c", "file", b"bye-id"),
            ("zz", "file", b"zz-id"),
            ("src/sub/", "directory", b"sub-id"),
            ("src/zz.c", "file", b"zzc-id"),
            ("src/sub/a", "file", b"a-id"),
            ("Makefile", "file", b"makefile-id"),
        ]:
            ie = inv.add_path(*args)
            ie.revision = b"revision"
            if args[1] == "file":
                ie.text_sha1 = osutils.sha_string(b"content\n")
                ie.text_size = len(b"content\n")
        return self.inv_to_test_inv(inv)


class TestInventoryCreateByApplyDelta(TestInventory):
    """A subset of the inventory delta application tests.

    See test_inv which has comprehensive delta application tests for
    inventories, dirstate, and repository based inventories.
    """

    def test_add(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta(
            [
                (None, "a", b"a-id", self.make_file(b"a-id", "a", b"tree-root")),
            ],
            b"new-test-rev",
        )
        self.assertEqual("a", inv.id2path(b"a-id"))

    def test_delete(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta(
            [
                (None, "a", b"a-id", self.make_file(b"a-id", "a", b"tree-root")),
            ],
            b"new-rev-1",
        )
        self.assertEqual("a", inv.id2path(b"a-id"))
        inv = inv.create_by_apply_delta(
            [
                ("a", None, b"a-id", None),
            ],
            b"new-rev-2",
        )
        self.assertRaises(errors.NoSuchId, inv.id2path, b"a-id")

    def test_rename(self):
        inv = self.make_init_inventory()
        inv = inv.create_by_apply_delta(
            [
                (None, "a", b"a-id", self.make_file(b"a-id", "a", b"tree-root")),
            ],
            b"new-rev-1",
        )
        self.assertEqual("a", inv.id2path(b"a-id"))
        a_ie = inv.get_entry(b"a-id")
        b_ie = self.make_file(a_ie.file_id, "b", a_ie.parent_id)
        inv = inv.create_by_apply_delta([("a", "b", b"a-id", b_ie)], b"new-rev-2")
        self.assertEqual("b", inv.id2path(b"a-id"))

    def test_illegal(self):
        # A file-id cannot appear in a delta more than once
        inv = self.make_init_inventory()
        self.assertRaises(
            errors.InconsistentDelta,
            inv.create_by_apply_delta,
            [
                (None, "a", b"id-1", self.make_file(b"id-1", "a", b"tree-root")),
                (None, "b", b"id-1", self.make_file(b"id-1", "b", b"tree-root")),
            ],
            b"new-rev-1",
        )


class TestInventoryReads(TestInventory):
    def test_is_root(self):
        """Ensure our root-checking code is accurate."""
        inv = self.make_init_inventory()
        self.assertTrue(inv.is_root(b"tree-root"))
        self.assertFalse(inv.is_root(b"booga"))
        ie = inv.get_entry(b"tree-root").copy()
        ie.file_id = b"booga"
        inv = inv.create_by_apply_delta(
            [("", None, b"tree-root", None), (None, "", b"booga", ie)], b"new-rev-2"
        )
        self.assertFalse(inv.is_root(b"TREE_ROOT"))
        self.assertTrue(inv.is_root(b"booga"))

    def test_ids(self):
        """Test detection of files within selected directories."""
        inv = inventory.Inventory(b"TREE_ROOT")
        inv.root.revision = b"revision"
        for args in [
            ("src", "directory", b"src-id"),
            ("doc", "directory", b"doc-id"),
            ("src/hello.c", "file"),
            ("src/bye.c", "file", b"bye-id"),
            ("Makefile", "file"),
        ]:
            ie = inv.add_path(*args)
            ie.revision = b"revision"
            if args[1] == "file":
                ie.text_sha1 = osutils.sha_string(b"content\n")
                ie.text_size = len(b"content\n")
        inv = self.inv_to_test_inv(inv)
        self.assertEqual(inv.path2id("src"), b"src-id")
        self.assertEqual(inv.path2id("src/bye.c"), b"bye-id")

    def test_get_entry_by_path_partial(self):
        inv = inventory.Inventory(b"TREE_ROOT")
        inv.root.revision = b"revision"
        for args in [
            ("src", "directory", b"src-id"),
            ("doc", "directory", b"doc-id"),
            ("src/hello.c", "file"),
            ("src/bye.c", "file", b"bye-id"),
            ("Makefile", "file"),
            ("external", "tree-reference", b"other-root"),
        ]:
            ie = inv.add_path(*args)
            ie.revision = b"revision"
            if args[1] == "file":
                ie.text_sha1 = osutils.sha_string(b"content\n")
                ie.text_size = len(b"content\n")
            if args[1] == "tree-reference":
                ie.reference_revision = b"reference"
        inv = self.inv_to_test_inv(inv)

        # Standard lookups
        ie, resolved, remaining = inv.get_entry_by_path_partial("")
        self.assertEqual((ie.file_id, resolved, remaining), (b"TREE_ROOT", [], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial("src")
        self.assertEqual((ie.file_id, resolved, remaining), (b"src-id", ["src"], []))
        ie, resolved, remaining = inv.get_entry_by_path_partial("src/bye.c")
        self.assertEqual(
            (ie.file_id, resolved, remaining), (b"bye-id", ["src", "bye.c"], [])
        )

        # Paths in the external tree
        ie, resolved, remaining = inv.get_entry_by_path_partial("external")
        self.assertEqual(
            (ie.file_id, resolved, remaining), (b"other-root", ["external"], [])
        )
        ie, resolved, remaining = inv.get_entry_by_path_partial("external/blah")
        self.assertEqual(
            (ie.file_id, resolved, remaining), (b"other-root", ["external"], ["blah"])
        )

        # Nonexistant paths
        ie, resolved, remaining = inv.get_entry_by_path_partial("foo.c")
        self.assertEqual((ie, resolved, remaining), (None, None, None))

    def test_non_directory_children(self):
        """Test path2id when a parent directory has no children."""
        inv = inventory.Inventory(b"tree-root")
        inv.add(self.make_file(b"file-id", "file", b"tree-root"))
        inv.add(self.make_link(b"link-id", "link", b"tree-root"))
        self.assertIs(None, inv.path2id("file/subfile"))
        self.assertIs(None, inv.path2id("link/subfile"))

    def test_is_unmodified(self):
        f1 = self.make_file(b"file-id", "file", b"tree-root")
        f1.revision = b"rev"
        self.assertTrue(f1.is_unmodified(f1))
        f2 = self.make_file(b"file-id", "file", b"tree-root")
        f2.revision = b"rev"
        self.assertTrue(f1.is_unmodified(f2))
        f3 = self.make_file(b"file-id", "file", b"tree-root")
        self.assertFalse(f1.is_unmodified(f3))
        f4 = self.make_file(b"file-id", "file", b"tree-root")
        f4.revision = b"rev1"
        self.assertFalse(f1.is_unmodified(f4))

    def test_iter_entries(self):
        inv = self.prepare_inv_with_nested_dirs()

        # Test all entries
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
                ("src/zz.c", b"zzc-id"),
                ("zz", b"zz-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries()],
        )

        # Test a subdirectory
        self.assertEqual(
            [
                ("bye.c", b"bye-id"),
                ("hello.c", b"hello-id"),
                ("sub", b"sub-id"),
                ("sub/a", b"a-id"),
                ("zz.c", b"zzc-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries(from_dir=b"src-id")],
        )

        # Test not recursing at the root level
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("zz", b"zz-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries(recursive=False)],
        )

        # Test not recursing at a subdirectory level
        self.assertEqual(
            [
                ("bye.c", b"bye-id"),
                ("hello.c", b"hello-id"),
                ("sub", b"sub-id"),
                ("zz.c", b"zzc-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries(from_dir=b"src-id", recursive=False)
            ],
        )

    def test_iter_just_entries(self):
        inv = self.prepare_inv_with_nested_dirs()
        self.assertEqual(
            [
                b"a-id",
                b"bye-id",
                b"doc-id",
                b"hello-id",
                b"makefile-id",
                b"src-id",
                b"sub-id",
                b"tree-root",
                b"zz-id",
                b"zzc-id",
            ],
            sorted([ie.file_id for ie in inv.iter_just_entries()]),
        )

    def test_iter_entries_by_dir(self):
        inv = self.prepare_inv_with_nested_dirs()
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("zz", b"zz-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/zz.c", b"zzc-id"),
                ("src/sub/a", b"a-id"),
            ],
            [(path, ie.file_id) for path, ie in inv.iter_entries_by_dir()],
        )
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("zz", b"zz-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/zz.c", b"zzc-id"),
                ("src/sub/a", b"a-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids=(
                        b"a-id",
                        b"zzc-id",
                        b"doc-id",
                        b"tree-root",
                        b"hello-id",
                        b"bye-id",
                        b"zz-id",
                        b"src-id",
                        b"makefile-id",
                        b"sub-id",
                    )
                )
            ],
        )

        self.assertEqual(
            [
                ("Makefile", b"makefile-id"),
                ("doc", b"doc-id"),
                ("zz", b"zz-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/zz.c", b"zzc-id"),
                ("src/sub/a", b"a-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids=(
                        b"a-id",
                        b"zzc-id",
                        b"doc-id",
                        b"hello-id",
                        b"bye-id",
                        b"zz-id",
                        b"makefile-id",
                    )
                )
            ],
        )

        self.assertEqual(
            [
                ("Makefile", b"makefile-id"),
                ("src/bye.c", b"bye-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids=(b"bye-id", b"makefile-id")
                )
            ],
        )

        self.assertEqual(
            [
                ("Makefile", b"makefile-id"),
                ("src/bye.c", b"bye-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(
                    specific_file_ids=(b"bye-id", b"makefile-id")
                )
            ],
        )

        self.assertEqual(
            [
                ("src/bye.c", b"bye-id"),
            ],
            [
                (path, ie.file_id)
                for path, ie in inv.iter_entries_by_dir(specific_file_ids=(b"bye-id",))
            ],
        )


class TestInventoryFiltering(TestInventory):
    def test_inv_filter_empty(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([])
        self.assertEqual(
            [
                ("", b"tree-root"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_files(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b"zz-id", b"hello-id", b"a-id"])
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("src", b"src-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
                ("zz", b"zz-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_dirs(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b"doc-id", b"sub-id"])
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("doc", b"doc-id"),
                ("src", b"src-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_files_and_dirs(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b"makefile-id", b"src-id"])
        self.assertEqual(
            [
                ("", b"tree-root"),
                ("Makefile", b"makefile-id"),
                ("src", b"src-id"),
                ("src/bye.c", b"bye-id"),
                ("src/hello.c", b"hello-id"),
                ("src/sub", b"sub-id"),
                ("src/sub/a", b"a-id"),
                ("src/zz.c", b"zzc-id"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )

    def test_inv_filter_entry_not_present(self):
        inv = self.prepare_inv_with_nested_dirs()
        new_inv = inv.filter([b"not-present-id"])
        self.assertEqual(
            [
                ("", b"tree-root"),
            ],
            [(path, ie.file_id) for path, ie in new_inv.iter_entries()],
        )



# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the generic Tree.walkdirs interface."""

import os

from breezy import tests
from breezy.mutabletree import MutableTree
from breezy.osutils import supports_symlinks
from breezy.tests.per_tree import TestCaseWithTree


class TestWalkdirs(TestCaseWithTree):
    def get_all_subdirs_expected(self, tree, symlinks):
        empty_dirs_present = tree.has_versioned_directories() or isinstance(
            tree, MutableTree
        )
        empty_dirs_are_versioned = tree.has_versioned_directories()
        dirblocks = {}

        dirblocks[""] = [
            ("0file", "0file", "file", None, "file"),
            ("1top-dir", "1top-dir", "directory", None, "directory"),
            ("2utf\u1234file", "2utf\u1234file", "file", None, "file"),
        ]

        dirblocks["1top-dir"] = [
            ("1top-dir/0file-in-1topdir", "0file-in-1topdir", "file", None, "file")
        ]
        if empty_dirs_present:
            dirblocks["1top-dir"].append(
                (
                    "1top-dir/1dir-in-1topdir",
                    "1dir-in-1topdir",
                    "directory",
                    None
                    if empty_dirs_are_versioned
                    else os.stat(tree.abspath("1top-dir/1dir-in-1topdir")),
                    "directory" if empty_dirs_are_versioned else None,
                )
            )
            dirblocks["1top-dir/1dir-in-1topdir"] = []
        if symlinks:
            dirblocks[""].append(("symlink", "symlink", "symlink", None, "symlink"))
        return [
            (path, sorted(entries)) for (path, entries) in sorted(dirblocks.items())
        ]

    def test_walkdir_root(self):
        tree = self.get_tree_with_subdirs_and_all_supported_content_types(
            supports_symlinks(self.test_dir)
        )
        with tree.lock_read():
            expected_dirblocks = self.get_all_subdirs_expected(
                tree, supports_symlinks(self.test_dir)
            )
            # test that its iterable by iterating
            result = []
            for dirinfo, block in tree.walkdirs():
                newblock = []
                for row in block:
                    if row[4] is not None:
                        newblock.append(row[0:3] + (None,) + row[4:])
                    else:
                        newblock.append(row)
                result.append((dirinfo, newblock))
        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            self.assertEqual(item, result[pos])
        self.assertEqual(len(expected_dirblocks), len(result))

    def test_walkdir_subtree(self):
        tree = self.get_tree_with_subdirs_and_all_supported_content_types(
            supports_symlinks(self.test_dir)
        )
        # test that its iterable by iterating
        result = []
        with tree.lock_read():
            expected_dirblocks = self.get_all_subdirs_expected(
                tree, supports_symlinks(self.test_dir)
            )[1:]
            for dirinfo, block in tree.walkdirs("1top-dir"):
                newblock = []
                for row in block:
                    if row[4] is not None:
                        newblock.append(row[0:3] + (None,) + row[4:])
                    else:
                        newblock.append(row)
                result.append((dirinfo, newblock))
        # check each return value for debugging ease.
        for pos, item in enumerate(expected_dirblocks):
            self.assertEqual(item, result[pos])
        self.assertEqual(len(expected_dirblocks), len(result))

    def test_walkdir_versioned_kind(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file", "tree/dir/"])
        work_tree.add(["file", "dir"])
        os.unlink("tree/file")
        os.rmdir("tree/dir")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        if not tree.supports_file_ids or tree.path2id("file") is None:
            raise tests.TestNotApplicable("Tree type cannot represent dangling ids.")
        expected = [
            (
                "",
                (
                    [("dir", "dir", "unknown", None, "directory")]
                    if tree.has_versioned_directories()
                    else []
                )
                + [("file", "file", "unknown", None, "file")],
            )
        ]
        if tree.has_versioned_directories():
            expected.append(("dir", []))
        self.assertEqual(expected, list(tree.walkdirs()))



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.pyutils."""

from breezy import branch, tests
from breezy.pyutils import calc_parent_name, get_named_object


class TestGetNamedObject(tests.TestCase):
    """Tests for get_named_object."""

    def test_module_only(self):
        import sys

        self.assertIs(sys, get_named_object("sys"))

    def test_dotted_module(self):
        self.assertIs(branch, get_named_object("breezy.branch"))

    def test_module_attr(self):
        self.assertIs(branch.Branch, get_named_object("breezy.branch", "Branch"))

    def test_dotted_attr(self):
        self.assertIs(
            branch.Branch.hooks, get_named_object("breezy.branch", "Branch.hooks")
        )

    def test_package(self):
        # breezy.tests is a package, not simply a module
        self.assertIs(tests, get_named_object("breezy.tests"))

    def test_package_attr(self):
        # breezy.tests is a package, not simply a module
        self.assertIs(tests.TestCase, get_named_object("breezy.tests", "TestCase"))

    def test_import_error(self):
        self.assertRaises(ImportError, get_named_object, "NO_SUCH_MODULE")

    def test_attribute_error(self):
        self.assertRaises(AttributeError, get_named_object, "sys", "NO_SUCH_ATTR")


class TestCalcParent_name(tests.TestCase):
    """Tests for calc_parent_name."""

    def test_dotted_member(self):
        self.assertEqual(
            ("mod_name", "attr1", "attr2"), calc_parent_name("mod_name", "attr1.attr2")
        )

    def test_undotted_member(self):
        self.assertEqual(
            ("mod_name", None, "attr1"), calc_parent_name("mod_name", "attr1")
        )

    def test_dotted_module_no_member(self):
        self.assertEqual(("mod", None, "sub_mod"), calc_parent_name("mod.sub_mod"))

    def test_undotted_module_no_member(self):
        err = self.assertRaises(AssertionError, calc_parent_name, "mod_name")
        self.assertEqual(
            "No parent object for top-level module 'mod_name'", err.args[0]
        )



# Copyright (C) 2007 by Jelmer Vernooij <jelmer@samba.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Couple of blackbox tests for the rewrite plugin."""

import os

from ....branch import Branch
from ....tests.blackbox import ExternalBase


class TestRebaseSimple(ExternalBase):
    def make_file(self, name, contents):
        with open(name, "w" + ("b" if isinstance(contents, bytes) else "")) as f:
            f.write(contents)

    def setUp(self):
        super().setUp()
        os.mkdir("main")
        os.chdir("main")
        self.run_bzr("init")
        self.make_file("hello", b"hi world")
        self.run_bzr("add")
        self.run_bzr("commit -m bla")
        self.run_bzr("branch . ../feature")

    def test_no_upstream_branch(self):
        self.run_bzr_error(["brz: ERROR: No upstream branch specified.\n"], "rebase")

    def test_notneeded(self):
        os.chdir("../feature")
        self.assertEqual("No revisions to rebase.\n", self.run_bzr("rebase ../main")[0])

    def test_custom_merge_type(self):
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        os.chdir("../feature")
        self.make_file("hoi", "my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        self.assertEqual("", self.run_bzr("rebase --lca ../main")[0])
        self.assertEqual("3\n", self.run_bzr("revno")[0])

    def test_notneeded_feature_ahead(self):
        os.chdir("../feature")
        self.make_file("barbla", "bloe")
        self.run_bzr("add")
        self.run_bzr("commit -m bloe")
        self.assertEqual("No revisions to rebase.\n", self.run_bzr("rebase ../main")[0])

    def test_notneeded_main_ahead(self):
        self.make_file("barbla", "bloe")
        self.run_bzr("add")
        self.run_bzr("commit -m bloe")
        os.chdir("../feature")
        self.assertEqual(
            "Base branch is descendant of current branch. Pulling instead.\n",
            self.run_bzr("rebase ../main")[0],
        )
        self.assertEqual(
            Branch.open("../feature").last_revision_info(),
            Branch.open("../main").last_revision_info(),
        )

    def test_no_pending_merges(self):
        self.run_bzr_error(
            ["brz: ERROR: No pending merges present.\n"], ["rebase", "--pending-merges"]
        )

    def test_pending_merges(self):
        os.chdir("..")
        self.build_tree_contents([("main/hello", "42")])
        self.run_bzr("add", working_dir="main")
        self.run_bzr("commit -m that main")
        self.build_tree_contents([("feature/hoi", "my data")])
        self.run_bzr("add", working_dir="feature")
        self.run_bzr("commit -m this feature")
        self.assertEqual(
            ("", " M  hello\nAll changes applied successfully.\n"),
            self.run_bzr("merge ../main", working_dir="feature"),
        )
        out, err = self.run_bzr("rebase --pending-merges", working_dir="feature")
        self.assertEqual("", out)
        self.assertContainsRe(err, "modified hello")
        self.assertEqual(("3\n", ""), self.run_bzr("revno", working_dir="feature"))

    def test_simple_success(self):
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        os.chdir("../feature")
        self.make_file("hoi", "my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        self.assertEqual("", self.run_bzr("rebase ../main")[0])
        self.assertEqual("3\n", self.run_bzr("revno")[0])

    def test_range(self):
        # commit mainline rev 2
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        # commit feature rev 2
        os.chdir("../feature")
        self.make_file("hoi", "my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        # commit feature rev 3
        self.make_file("hooi", "your data")
        self.run_bzr("add")
        self.run_bzr("commit -m that")
        # commit feature rev 4
        self.make_file("hoooi", "someone else's data")
        self.run_bzr("add")
        self.run_bzr("commit -m these")
        # pick up just rev 2 and 3 and discard 4 from feature
        self.assertEqual("", self.run_bzr("rebase -r2..3 ../main")[0])
        # our rev 2 is now rev3 and 3 is now rev4:
        self.assertEqual("4\n", self.run_bzr("revno")[0])
        # content added from our old revisions 4 should be gone.
        self.assertPathDoesNotExist("hoooi")

    def test_range_open_end(self):
        # commit mainline rev 2
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        # commit feature rev 2
        os.chdir("../feature")
        self.make_file("hoi", "my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        # commit feature rev 3
        self.make_file("hooi", "your data")
        self.run_bzr("add")
        self.run_bzr("commit -m that")
        # commit feature rev 4
        self.make_file("hoooi", "someone else's data")
        self.run_bzr("add")
        self.run_bzr("commit -m these")
        # rebase only rev 4 onto main
        self.assertEqual("", self.run_bzr("rebase -r4.. ../main")[0])
        # should only get rev 3 (our old 2 and 3 are gone)
        self.assertEqual("3\n", self.run_bzr("revno")[0])
        self.assertPathDoesNotExist("hoi")
        self.assertPathDoesNotExist("hooi")
        branch = Branch.open(".")
        self.assertEqual(
            "these", branch.repository.get_revision(branch.last_revision()).message
        )
        self.assertPathExists("hoooi")

    def test_conflicting(self):
        # commit mainline rev 2
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        # commit feature rev 2 changing hello differently
        os.chdir("../feature")
        self.make_file("hello", "other data")
        self.run_bzr("commit -m this")
        self.run_bzr_error(
            [
                "Text conflict in hello\n1 conflicts encountered.\nbrz: ERROR: A conflict occurred replaying a commit. Resolve the conflict and run 'brz rebase-continue' or run 'brz rebase-abort'.",
            ],
            ["rebase", "../main"],
        )

    def test_conflicting_abort(self):
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        os.chdir("../feature")
        self.make_file("hello", "other data")
        self.run_bzr("commit -m this")
        old_log = self.run_bzr("log")[0]
        self.run_bzr_error(
            [
                "Text conflict in hello\n1 conflicts encountered.\nbrz: ERROR: A conflict occurred replaying a commit. Resolve the conflict and run 'brz rebase-continue' or run 'brz rebase-abort'.\n"
            ],
            ["rebase", "../main"],
        )
        self.assertEqual("", self.run_bzr("rebase-abort")[0])
        self.assertEqual(old_log, self.run_bzr("log")[0])

    def test_conflicting_continue(self):
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        os.chdir("../feature")
        self.make_file("hello", "other data")
        self.run_bzr("commit -m this")
        self.run_bzr_error(
            [
                "Text conflict in hello\n1 conflicts encountered.\nbrz: ERROR: A conflict occurred replaying a commit. Resolve the conflict and run 'brz rebase-continue' or run 'brz rebase-abort'.\n"
            ],
            ["rebase", "../main"],
        )
        self.run_bzr("resolved hello")
        self.assertEqual("", self.run_bzr("rebase-continue")[0])
        self.assertEqual("3\n", self.run_bzr("revno")[0])

    def test_continue_nothing(self):
        self.run_bzr_error(["brz: ERROR: No rebase to continue"], ["rebase-continue"])

    def test_abort_nothing(self):
        self.run_bzr_error(["brz: ERROR: No rebase to abort"], ["rebase-abort"])

    def test_todo_nothing(self):
        self.run_bzr_error(["brz: ERROR: No rebase in progress"], ["rebase-todo"])

    def test_onto(self):
        self.make_file("hello", "42")
        self.run_bzr("add")
        self.run_bzr("commit -m that")
        self.make_file("other", "43")
        self.run_bzr("add")
        self.run_bzr("commit -m that_other")
        os.chdir("../feature")
        self.make_file("hoi", "my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        self.assertEqual("", self.run_bzr("rebase --onto -2 ../main")[0])
        self.assertEqual("3\n", self.run_bzr("revno")[0])

    def test_unrelated(self):
        os.chdir("..")
        os.mkdir("unrelated")
        os.chdir("unrelated")
        self.run_bzr("init")
        self.make_file("hello", "hi world")
        self.run_bzr("add")
        self.run_bzr("commit -m x")
        self.run_bzr_error(
            ["brz: ERROR: Branches have no common ancestor, and no merge base.*"],
            ["rebase", "../main"],
        )

    def test_verbose(self):
        self.make_file("hello", "42")
        self.run_bzr("commit -m that")
        os.chdir("../feature")
        self.make_file("hoi", "my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        out, err = self.run_bzr("rebase -v ../main")
        self.assertContainsRe(err, "1 revisions will be rebased:")
        self.assertEqual("", out)
        self.assertEqual("3\n", self.run_bzr("revno")[0])

    def test_useless_merge(self):
        self.make_file("bar", "42")
        self.run_bzr("add")
        self.run_bzr("commit -m that")
        os.chdir("../feature")
        self.make_file("hello", "my data")
        self.run_bzr("commit -m this")
        self.run_bzr("merge")
        self.run_bzr("commit -m merge")
        self.run_bzr("rebase")

    def strip_last_revid_part(self, revid):
        """Assume revid is a revid in the default form, and strip the part
        which would be random.
        """
        return revid[: revid.rindex(b"-")]

    def test_always_rebase_merges(self):
        trunk = self.make_branch_and_tree("trunk")
        trunk.commit("base")
        feature2 = trunk.controldir.sprout("feature2").open_workingtree()
        revid2 = feature2.commit("change")
        feature = trunk.controldir.sprout("feature").open_workingtree()
        feature.commit("change")
        feature.merge_from_branch(feature2.branch)
        feature.commit("merge")
        feature.commit("change2")
        trunk.commit("additional upstream change")
        self.run_bzr("rebase --always-rebase-merges ../trunk", working_dir="feature")
        # second revision back should be a merge of feature2
        repo = feature.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        tip = feature.last_revision()
        merge_id = repo.get_graph().get_parent_map([tip])[tip][0]
        merge_parents = repo.get_graph().get_parent_map([merge_id])[merge_id]
        self.assertEqual(
            self.strip_last_revid_part(revid2),
            self.strip_last_revid_part(merge_parents[1]),
        )

    def test_rebase_merge(self):
        trunk = self.make_branch_and_tree("trunk")
        trunk.commit("base")
        feature2 = trunk.controldir.sprout("feature2").open_workingtree()
        revid2 = feature2.commit("change")
        feature = trunk.controldir.sprout("feature").open_workingtree()
        feature.commit("change")
        feature.merge_from_branch(feature2.branch)
        feature.commit("merge")
        feature.commit("change2")
        trunk.commit("additional upstream change")
        self.run_bzr("rebase ../trunk", working_dir="feature")
        # second revision back should be a merge of feature2
        repo = feature.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        tip = feature.last_revision()
        merge_id = repo.get_graph().get_parent_map([tip])[tip][0]
        merge_parents = repo.get_graph().get_parent_map([merge_id])[merge_id]
        self.assertEqual(
            self.strip_last_revid_part(revid2),
            self.strip_last_revid_part(merge_parents[1]),
        )

    def test_directory(self):
        self.make_file("test_directory1", "testing non-current directories")
        self.run_bzr("add")
        self.run_bzr("commit -m blah")
        os.chdir("../feature")
        self.make_file("test_directory2", "testing non-current directories")
        self.run_bzr("add")
        self.run_bzr("commit -m blah")
        os.chdir("..")
        self.assertEqual("", self.run_bzr("rebase -d feature main")[0])


class ReplayTests(ExternalBase):
    def test_replay(self):
        os.mkdir("main")
        os.chdir("main")
        self.run_bzr("init")
        with open("bar", "w") as f:
            f.write("42")
        self.run_bzr("add")
        self.run_bzr("commit -m that")
        os.mkdir("../feature")
        os.chdir("../feature")
        self.run_bzr("init")
        branch = Branch.open(".")
        with open("hello", "w") as f:
            f.write("my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        self.assertEqual(1, branch.revno())
        self.run_bzr("replay -r1 ../main")
        self.assertEqual(2, branch.revno())
        self.assertTrue(os.path.exists("bar"))

    def test_replay_open_range(self):
        os.mkdir("main")
        os.chdir("main")
        self.run_bzr("init")
        with open("bar", "w") as f:
            f.write("42")
        self.run_bzr("add")
        self.run_bzr("commit -m that")
        with open("bar", "w") as f:
            f.write("84")
        self.run_bzr("commit -m blathat")
        os.mkdir("../feature")
        os.chdir("../feature")
        self.run_bzr("init")
        branch = Branch.open(".")
        with open("hello", "w") as f:
            f.write("my data")
        self.run_bzr("add")
        self.run_bzr("commit -m this")
        self.assertEqual(1, branch.revno())
        self.run_bzr("replay -r1.. ../main")
        self.assertEqual(3, branch.revno())
        self.assertTrue(os.path.exists("bar"))



# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from breezy import controldir
from breezy.tests import TestCaseWithTransport


class TestReference(TestCaseWithTransport):
    def get_default_format(self):
        return controldir.format_registry.make_controldir("2a")

    def test_no_args_lists(self):
        tree = self.make_branch_and_tree("branch")
        tree.add_reference(self.make_branch_and_tree("branch/path"))
        tree.add_reference(self.make_branch_and_tree("branch/lath"))
        tree.set_reference_info("path", "http://example.org")
        tree.set_reference_info("lath", "http://example.org/2")
        out, err = self.run_bzr("reference", working_dir="branch")
        lines = out.splitlines()
        self.assertEqual("lath http://example.org/2", lines[0])
        self.assertEqual("path http://example.org", lines[1])

    def make_tree_with_reference(self):
        tree = self.make_branch_and_tree("tree")
        subtree = self.make_branch_and_tree("tree/newpath")
        tree.add_reference(subtree)
        tree.set_reference_info("newpath", "http://example.org")
        tree.commit("add reference")
        return tree

    def test_uses_working_tree_location(self):
        self.make_tree_with_reference()
        out, err = self.run_bzr("reference", working_dir="tree")
        self.assertContainsRe(out, "newpath http://example.org\n")

    def test_uses_basis_tree_location(self):
        tree = self.make_tree_with_reference()
        tree.controldir.destroy_workingtree()
        out, err = self.run_bzr("reference", working_dir="tree")
        self.assertContainsRe(out, "newpath http://example.org\n")

    def test_one_arg_displays(self):
        self.make_tree_with_reference()
        out, err = self.run_bzr("reference newpath", working_dir="tree")
        self.assertEqual("newpath http://example.org\n", out)

    def test_one_arg_uses_containing_tree(self):
        self.make_tree_with_reference()
        out, err = self.run_bzr("reference -d tree newpath")
        self.assertEqual("newpath http://example.org\n", out)

    def test_two_args_sets(self):
        tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file"])
        tree.add("file")
        out, err = self.run_bzr("reference -d tree file http://example.org")
        location = tree.get_reference_info("file")
        self.assertEqual("http://example.org", location)
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_missing_file(self):
        self.make_branch_and_tree("tree")
        out, err = self.run_bzr(
            "reference file http://example.org", working_dir="tree", retcode=3
        )
        self.assertEqual("brz: ERROR: file is not versioned.\n", err)

    def test_missing_file_forced(self):
        tree = self.make_branch_and_tree("tree")
        tree.add_reference(self.make_branch_and_tree("tree/file"))
        out, err = self.run_bzr(
            "reference --force-unversioned file http://example.org", working_dir="tree"
        )
        location = tree.get_reference_info("file")
        self.assertEqual("http://example.org", location)
        self.assertEqual("", out)
        self.assertEqual("", err)



# Copyright (C) 2010 by Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import subprocess
import sys

import breezy
from breezy import commands, osutils, tests
from breezy.plugins.bash_completion.bashcomp import *
from breezy.tests import features


class BashCompletionMixin:
    """Component for testing execution of a bash completion script."""

    _test_needs_features = [features.bash_feature]
    script = None

    def complete(self, words, cword=-1):
        """Perform a bash completion.

        :param words: a list of words representing the current command.
        :param cword: the current word to complete, defaults to the last one.
        """
        if self.script is None:
            self.script = self.get_script()
        env = dict(os.environ)
        env["PYTHONPATH"] = ":".join(sys.path)
        proc = subprocess.Popen(
            [features.bash_feature.path, "--noprofile"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
        if cword < 0:
            cword = len(words) + cword
        encoding = osutils.get_user_encoding()
        input = b"%s\n" % self.script.encode(encoding)
        input += b"COMP_WORDS=( %s )\n" % b" ".join(
            [b"'" + w.replace("'", "'\\''").encode(encoding) + b"'" for w in words]
        )
        input += b"COMP_CWORD=%d\n" % cword
        input += b"%s\n" % getattr(self, "script_name", "_brz").encode(encoding)
        input += b"echo ${#COMPREPLY[*]}\n"
        input += b"IFS=$'\\n'\n"
        input += b'echo "${COMPREPLY[*]}"\n'
        (out, err) = proc.communicate(input)
        errlines = [
            line for line in err.splitlines() if not line.startswith(b"brz: warning: ")
        ]
        if errlines != []:
            raise AssertionError("Unexpected error message:\n{}".format(err))
        self.assertEqual(b"", b"".join(errlines), "No messages to standard error")
        # import sys
        # print >>sys.stdout, '---\n%s\n---\n%s\n---\n' % (input, out)
        lines = out.split(b"\n")
        nlines = int(lines[0])
        del lines[0]
        self.assertEqual(b"", lines[-1], "Newline at end")
        del lines[-1]
        if nlines == 0 and len(lines) == 1 and lines[0] == b"":
            del lines[0]
        self.assertEqual(nlines, len(lines), "No newlines in generated words")
        self.completion_result = {l.decode(encoding) for l in lines}
        return self.completion_result

    def assertCompletionEquals(self, *words):
        self.assertEqual(set(words), self.completion_result)

    def assertCompletionContains(self, *words):
        missing = set(words) - self.completion_result
        if missing:
            raise AssertionError(
                "Completion should contain {!r} but it has {!r}".format(
                    missing, self.completion_result
                )
            )

    def assertCompletionOmits(self, *words):
        surplus = set(words) & self.completion_result
        if surplus:
            raise AssertionError(
                "Completion should omit {!r} but it has {!r}".format(
                    surplus, self.completion_result
                )
            )

    def get_script(self):
        commands.install_bzr_command_hooks()
        dc = DataCollector()
        data = dc.collect()
        cg = BashCodeGen(data)
        res = cg.function()
        return res


class TestBashCompletion(tests.TestCase, BashCompletionMixin):
    """Test bash completions that don't execute brz."""

    def test_simple_scipt(self):
        """Ensure that the test harness works as expected."""
        self.script = """
_brz() {
    COMPREPLY=()
    # add all words in reverse order, with some markup around them
    for ((i = ${#COMP_WORDS[@]}; i > 0; --i)); do
        COMPREPLY+=( "-${COMP_WORDS[i-1]}+" )
    done
    # and append the current word
    COMPREPLY+=( "+${COMP_WORDS[COMP_CWORD]}-" )
}
"""
        self.complete(["foo", '"bar', "'baz"], cword=1)
        self.assertCompletionEquals("-'baz+", '-"bar+', "-foo+", '+"bar-')

    def test_cmd_ini(self):
        self.complete(["brz", "ini"])
        self.assertCompletionContains(
            "init", "init-shared-repo", "init-shared-repository"
        )
        self.assertCompletionOmits("commit")

    def test_init_opts(self):
        self.complete(["brz", "init", "-"])
        self.assertCompletionContains("-h", "--format=2a")

    def test_global_opts(self):
        self.complete(["brz", "-", "init"], cword=1)
        self.assertCompletionContains("--no-plugins", "--builtin")

    def test_commit_dashm(self):
        self.complete(["brz", "commit", "-m"])
        self.assertCompletionEquals("-m")

    def test_status_negated(self):
        self.complete(["brz", "status", "--n"])
        self.assertCompletionContains("--no-versioned", "--no-verbose")

    def test_init_format_any(self):
        self.complete(["brz", "init", "--format", "=", "directory"], cword=3)
        self.assertCompletionContains("1.9", "2a")

    def test_init_format_2(self):
        self.complete(["brz", "init", "--format", "=", "2", "directory"], cword=4)
        self.assertCompletionContains("2a")
        self.assertCompletionOmits("1.9")


class TestBashCompletionInvoking(tests.TestCaseWithTransport, BashCompletionMixin):
    """Test bash completions that might execute brz.

    Only the syntax ``$(brz ...`` is supported so far. The brz command
    will be replaced by the brz instance running this selftest.
    """

    def setUp(self):
        super().setUp()
        if sys.platform == "win32":
            raise tests.KnownFailure("see bug #709104, completion is broken on windows")

    def get_script(self):
        s = super().get_script()
        s = s.replace("$(brz ", "$({} ".format(" ".join(self.get_brz_command())))
        s = s.replace("2>/dev/null", "")
        return s

    def test_revspec_tag_all(self):
        self.requireFeature(features.sed_feature)
        wt = self.make_branch_and_tree(".")
        wt.branch.tags.set_tag("tag1", b"null:")
        wt.branch.tags.set_tag("tag2", b"null:")
        wt.branch.tags.set_tag("3tag", b"null:")
        self.complete(["brz", "log", "-r", "tag", ":"])
        self.assertCompletionEquals("tag1", "tag2", "3tag")

    def test_revspec_tag_prefix(self):
        self.requireFeature(features.sed_feature)
        wt = self.make_branch_and_tree(".")
        wt.branch.tags.set_tag("tag1", b"null:")
        wt.branch.tags.set_tag("tag2", b"null:")
        wt.branch.tags.set_tag("3tag", b"null:")
        self.complete(["brz", "log", "-r", "tag", ":", "t"])
        self.assertCompletionEquals("tag1", "tag2")

    def test_revspec_tag_spaces(self):
        self.requireFeature(features.sed_feature)
        wt = self.make_branch_and_tree(".")
        wt.branch.tags.set_tag("tag with spaces", b"null:")
        self.complete(["brz", "log", "-r", "tag", ":", "t"])
        self.assertCompletionEquals(r"tag\ with\ spaces")
        self.complete(["brz", "log", "-r", '"tag:t'])
        self.assertCompletionEquals("tag:tag with spaces")
        self.complete(["brz", "log", "-r", "'tag:t"])
        self.assertCompletionEquals("tag:tag with spaces")

    def test_revspec_tag_endrange(self):
        self.requireFeature(features.sed_feature)
        wt = self.make_branch_and_tree(".")
        wt.branch.tags.set_tag("tag1", b"null:")
        wt.branch.tags.set_tag("tag2", b"null:")
        self.complete(["brz", "log", "-r", "3..tag", ":", "t"])
        self.assertCompletionEquals("tag1", "tag2")
        self.complete(["brz", "log", "-r", '"3..tag:t'])
        self.assertCompletionEquals("3..tag:tag1", "3..tag:tag2")
        self.complete(["brz", "log", "-r", "'3..tag:t"])
        self.assertCompletionEquals("3..tag:tag1", "3..tag:tag2")


class TestBashCodeGen(tests.TestCase):
    def test_command_names(self):
        data = CompletionData()
        bar = CommandData("bar")
        bar.aliases.append("baz")
        data.commands.append(bar)
        data.commands.append(CommandData("foo"))
        cg = BashCodeGen(data)
        self.assertEqual("bar baz foo", cg.command_names())

    def test_debug_output(self):
        data = CompletionData()
        self.assertEqual("", BashCodeGen(data, debug=False).debug_output())
        self.assertTrue(BashCodeGen(data, debug=True).debug_output())

    def test_brz_version(self):
        data = CompletionData()
        cg = BashCodeGen(data)
        self.assertEqual("{}.".format(breezy.version_string), cg.brz_version())
        data.plugins["foo"] = PluginData("foo", "1.0")
        data.plugins["bar"] = PluginData("bar", "2.0")
        cg = BashCodeGen(data)
        self.assertEqual(
            """\
{} and the following plugins:
# bar 2.0
# foo 1.0""".format(breezy.version_string),
            cg.brz_version(),
        )

    def test_global_options(self):
        data = CompletionData()
        data.global_options.add("--foo")
        data.global_options.add("--bar")
        cg = BashCodeGen(data)
        self.assertEqual("--bar --foo", cg.global_options())

    def test_command_cases(self):
        data = CompletionData()
        bar = CommandData("bar")
        bar.aliases.append("baz")
        bar.options.append(OptionData("--opt"))
        data.commands.append(bar)
        data.commands.append(CommandData("foo"))
        cg = BashCodeGen(data)
        self.assertEqualDiff(
            """\
\tbar|baz)
\t\tcmdOpts=( --opt )
\t\t;;
\tfoo)
\t\tcmdOpts=(  )
\t\t;;
""",
            cg.command_cases(),
        )

    def test_command_case(self):
        cmd = CommandData("cmd")
        cmd.plugin = PluginData("plugger", "1.0")
        bar = OptionData("--bar")
        bar.registry_keys = ["that", "this"]
        bar.error_messages.append("Some error message")
        cmd.options.append(bar)
        cmd.options.append(OptionData("--foo"))
        data = CompletionData()
        data.commands.append(cmd)
        cg = BashCodeGen(data)
        self.assertEqualDiff(
            """\
\tcmd)
\t\t# plugin "plugger 1.0"
\t\t# Some error message
\t\tcmdOpts=( --bar=that --bar=this --foo )
\t\tcase $curOpt in
\t\t\t--bar) optEnums=( that this ) ;;
\t\tesac
\t\t;;
""",
            cg.command_case(cmd),
        )


class TestDataCollector(tests.TestCase):
    def setUp(self):
        super().setUp()
        commands.install_bzr_command_hooks()

    def test_global_options(self):
        dc = DataCollector()
        dc.global_options()
        self.assertSubset(["--no-plugins", "--builtin"], dc.data.global_options)

    def test_commands(self):
        dc = DataCollector()
        dc.commands()
        self.assertSubset(
            ["init", "init-shared-repo", "init-shared-repository"],
            dc.data.all_command_aliases(),
        )

    def test_commands_from_plugins(self):
        dc = DataCollector()
        dc.commands()
        self.assertSubset(["bash-completion"], dc.data.all_command_aliases())

    def test_commit_dashm(self):
        dc = DataCollector()
        cmd = dc.command("commit")
        self.assertSubset(["-m"], [str(o) for o in cmd.options])

    def test_status_negated(self):
        dc = DataCollector()
        cmd = dc.command("status")
        self.assertSubset(
            ["--no-versioned", "--no-verbose"], [str(o) for o in cmd.options]
        )

    def test_init_format(self):
        dc = DataCollector()
        cmd = dc.command("init")
        for opt in cmd.options:
            if opt.name == "--format":
                self.assertSubset(["2a"], opt.registry_keys)
                return
        raise AssertionError("Option --format not found")


class BlackboxTests(tests.TestCaseWithMemoryTransport):
    def test_bash_completion(self):
        self.run_bzr("bash-completion", encoding="utf-8")



#
# Breezy documentation build configuration file, created by
# sphinx-quickstart on Tue Jul 21 17:04:52 2009.
#
# This file is execfile()d with the current directory set to its containing dir.

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path = [os.path.abspath("../.."), os.path.abspath("api")] + sys.path

# Most of the configuration for Breezy docs is defined here ...
from breezy.doc_generate.conf import *

## Configuration specific to this site ##

# The locale code for this documentation set
brz_locale = "en"

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "Developer Document Catalog ({})".format(release)

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {'index': 'index.html'}

# Output file base name for HTML help builder.
htmlhelp_basename = "brz-developers"

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
brz_documents = [
    (
        "HACKING",
        "brz-en-developer-guide",
        "Breezy Developer Guide",
        "Breezy Developers",
        "manual",
    ),
    (
        "testing",
        "brz-en-testing-guide",
        "Breezy Testing Guide",
        "Breezy Developers",
        "manual",
    ),
    (
        "overview",
        "brz-en-architecture-overview",
        "Breezy Architecture Overview",
        "Breezy Developers",
        "howto",
    ),
    (
        "integration",
        "brz-en-integration-guide",
        "Breezy Integration Guide",
        "Breezy Developers",
        "howto",
    ),
]

latex_documents = [
    (start, target + ".tex", title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
]

texinfo_documents = [
    (start, target, title, author, doc_class)
    for start, target, title, author, doc_class in brz_documents
]

# List of documents that shouldn't be included in the build.
# Note: Maybe some of them *ought* to be linked in somewhere?
unused_docs = [
    "add",
    "annotate",
    "bundle-creation",
    "bundle-format4",
    "check",
    "commit",
    "diff",
    "directory-fingerprints",
    "gc",
    "index-plain",
    "incremental-push-pull",
    "initial-push-pull",
    "merge-scaling",
    "missing",
    "performance-roadmap-rationale",
    "performance-use-case-analysis",
    "planned-change-integration",
    "planned-performance-changes",
    "revert",
    "status",
    "uncommit",
]



from ...tests import TestCaseWithTransport


class TestBlackbox(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        wt = self.make_branch_and_tree(".")
        self.build_tree_contents([("foo.c", "#include <stdio.h>\n")])
        wt.add("foo.c")
        wt.commit(message="1", committer="Fero <fero@example.com>", rev_id=b"1")
        wt.commit(message="2", committer="Fero <fero@example.com>", rev_id=b"2")
        wt.commit(message="3", committer="Jano <jano@example.com>", rev_id=b"3")
        wt.commit(
            message="4",
            committer="Jano <jano@example.com>",
            authors=["Vinco <vinco@example.com>"],
            rev_id=b"4",
        )
        wt.commit(message="5", committer="Ferko <fero@example.com>", rev_id=b"5")

    def test_stats(self):
        (out, err) = self.run_bzr("stats")
        self.assertEqual(
            out,
            """\
   3 Fero <fero@example.com>
     Other names:
        2 Fero
        1 Ferko
   1 Vinco <vinco@example.com>
   1 Jano <jano@example.com>
""",
        )

    def test_credits(self):
        (out, err) = self.run_bzr("credits")
        self.assertEqual(
            out,
            """\
Code:
Fero <fero@example.com>

""",
        )



# Copyright (C) 2005, 2006, 2007, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from io import StringIO

from .. import add, errors, tests
from ..bzr import inventory


class AddCustomIDAction(add.AddAction):
    def __call__(self, inv, parent_ie, path, kind):
        # The first part just logs if appropriate
        # Now generate a custom id
        file_id = (kind + "-" + path.replace("/", "%")).encode("utf-8")
        if self.should_print:
            self._to_file.write(
                "added {} with id {}\n".format(path, file_id.decode("utf-8"))
            )
        return file_id


class TestAddFrom(tests.TestCaseWithTransport):
    """Tests for AddFromBaseAction."""

    def make_base_tree(self):
        self.base_tree = self.make_branch_and_tree("base")
        self.build_tree(
            [
                "base/a",
                "base/b",
                "base/dir/",
                "base/dir/a",
                "base/dir/subdir/",
                "base/dir/subdir/b",
            ]
        )
        self.base_tree.add(["a", "b", "dir", "dir/a", "dir/subdir", "dir/subdir/b"])
        self.base_tree.commit("creating initial tree.")

    def add_helper(self, base_tree, base_path, new_tree, file_list, should_print=False):
        to_file = StringIO()
        base_tree.lock_read()
        try:
            new_tree.lock_write()
            try:
                action = add.AddFromBaseAction(
                    base_tree, base_path, to_file=to_file, should_print=should_print
                )
                new_tree.smart_add(file_list, action=action)
            finally:
                new_tree.unlock()
        finally:
            base_tree.unlock()
        return to_file.getvalue()

    def test_copy_all(self):
        self.make_base_tree()
        new_tree = self.make_branch_and_tree("new")
        files = [
            "a",
            "b",
            "dir/",
            "dir/a",
            "dir/subdir/",
            "dir/subdir/b",
        ]
        self.build_tree(["new/" + fn for fn in files])
        self.add_helper(self.base_tree, "", new_tree, ["new"])

        for fn in files:
            base_file_id = self.base_tree.path2id(fn)
            new_file_id = new_tree.path2id(fn)
            self.assertEqual(base_file_id, new_file_id)

    def test_copy_from_dir(self):
        self.make_base_tree()
        new_tree = self.make_branch_and_tree("new")

        self.build_tree(
            ["new/a", "new/b", "new/c", "new/subdir/", "new/subdir/b", "new/subdir/d"]
        )
        new_tree.set_root_id(self.base_tree.path2id(""))
        self.add_helper(self.base_tree, "dir", new_tree, ["new"])

        # We know 'a' and 'b' exist in the root, and they are being added
        # in a new 'root'. Since ROOT ids have been set as the same, we will
        # use those ids
        self.assertEqual(self.base_tree.path2id("a"), new_tree.path2id("a"))
        self.assertEqual(self.base_tree.path2id("b"), new_tree.path2id("b"))

        # Because we specified 'dir/' as the base path, and we have
        # nothing named 'subdir' in the base tree, we should grab the
        # ids from there
        self.assertEqual(
            self.base_tree.path2id("dir/subdir"), new_tree.path2id("subdir")
        )
        self.assertEqual(
            self.base_tree.path2id("dir/subdir/b"), new_tree.path2id("subdir/b")
        )

        # These should get newly generated ids
        c_id = new_tree.path2id("c")
        self.assertNotEqual(None, c_id)
        self.base_tree.lock_read()
        self.addCleanup(self.base_tree.unlock)
        self.assertRaises(errors.NoSuchId, self.base_tree.id2path, c_id)

        d_id = new_tree.path2id("subdir/d")
        self.assertNotEqual(None, d_id)
        self.assertRaises(errors.NoSuchId, self.base_tree.id2path, d_id)

    def test_copy_existing_dir(self):
        self.make_base_tree()
        new_tree = self.make_branch_and_tree("new")
        self.build_tree(["new/subby/", "new/subby/a", "new/subby/b"])

        subdir_file_id = self.base_tree.path2id("dir/subdir")
        new_tree.add(["subby"], ids=[subdir_file_id])
        self.add_helper(self.base_tree, "", new_tree, ["new"])
        # Because 'subby' already points to subdir, we should add
        # 'b' with the same id
        self.assertEqual(
            self.base_tree.path2id("dir/subdir/b"), new_tree.path2id("subby/b")
        )

        # 'subby/a' should be added with a new id because there is no
        # matching path or child of 'subby'.
        a_id = new_tree.path2id("subby/a")
        self.assertNotEqual(None, a_id)
        self.base_tree.lock_read()
        self.addCleanup(self.base_tree.unlock)
        self.assertRaises(errors.NoSuchId, self.base_tree.id2path, a_id)


class TestAddActions(tests.TestCase):
    def test_quiet(self):
        self.run_action("")

    def test__print(self):
        self.run_action("adding path\n")

    def run_action(self, output):
        inv = inventory.Inventory()
        stdout = StringIO()
        action = add.AddAction(to_file=stdout, should_print=bool(output))

        self.apply_redirected(None, stdout, None, action, inv, None, "path", "file")
        self.assertEqual(stdout.getvalue(), output)



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Weave formats.

These were formats present in pre-1.0 version of Bazaar.
"""

# Since we are a built-in plugin we share the breezy version
from ... import branch as _mod_branch
from ... import (
    controldir,
    version_info,  # noqa: F401
)
from ... import repository as _mod_repository
from ... import workingtree as _mod_workingtree
from ...bzr import BzrProber, register_metadir, serializer

# Pre-0.8 formats that don't have a disk format string (because they are
# versioned by the matching control directory). We use the control directories
# disk format string as a key for the network_name because they meet the
# constraints (simple string, unique, immutable).
_mod_repository.network_format_registry.register_lazy(
    b"Bazaar-NG branch, format 5\n",
    "breezy.plugins.weave_fmt.repository",
    "RepositoryFormat5",
)
_mod_repository.network_format_registry.register_lazy(
    b"Bazaar-NG branch, format 6\n",
    "breezy.plugins.weave_fmt.repository",
    "RepositoryFormat6",
)

# weave formats which has no format string and are not discoverable or independently
# creatable on disk, so are not registered in format_registry.  They're
# all in breezy.plugins.weave_fmt.repository now.  When an instance of one of these is
# needed, it's constructed directly by the BzrDir.  Non-native formats where
# the repository is not separately opened are similar.

_mod_repository.format_registry.register_lazy(
    b"Bazaar-NG Repository format 7",
    "breezy.plugins.weave_fmt.repository",
    "RepositoryFormat7",
)

_mod_repository.format_registry.register_extra_lazy(
    "breezy.plugins.weave_fmt.repository", "RepositoryFormat4"
)
_mod_repository.format_registry.register_extra_lazy(
    "breezy.plugins.weave_fmt.repository", "RepositoryFormat5"
)
_mod_repository.format_registry.register_extra_lazy(
    "breezy.plugins.weave_fmt.repository", "RepositoryFormat6"
)


# The pre-0.8 formats have their repository format network name registered in
# repository.py. MetaDir formats have their repository format network name
# inferred from their disk format string.
controldir.format_registry.register_lazy(
    "weave",
    "breezy.plugins.weave_fmt.bzrdir",
    "BzrDirFormat6",
    "Pre-0.8 format.  Slower than knit and does not"
    " support checkouts or shared repositories.",
    hidden=True,
    deprecated=True,
)
register_metadir(
    controldir.format_registry,
    "metaweave",
    "breezy.plugins.weave_fmt.repository.RepositoryFormat7",
    "Transitional format in 0.8.  Slower than knit.",
    branch_format="breezy.bzr.fullhistory.BzrBranchFormat5",
    tree_format="breezy.bzr.workingtree_3.WorkingTreeFormat3",
    hidden=True,
    deprecated=True,
)


BzrProber.formats.register_lazy(
    b"Bazaar-NG branch, format 0.0.4\n",
    "breezy.plugins.weave_fmt.bzrdir",
    "BzrDirFormat4",
)
BzrProber.formats.register_lazy(
    b"Bazaar-NG branch, format 5\n", "breezy.plugins.weave_fmt.bzrdir", "BzrDirFormat5"
)
BzrProber.formats.register_lazy(
    b"Bazaar-NG branch, format 6\n", "breezy.plugins.weave_fmt.bzrdir", "BzrDirFormat6"
)


_mod_branch.format_registry.register_extra_lazy(
    "breezy.plugins.weave_fmt.branch", "BzrBranchFormat4"
)
_mod_branch.network_format_registry.register_lazy(
    b"Bazaar-NG branch, format 6\n",
    "breezy.plugins.weave_fmt.branch",
    "BzrBranchFormat4",
)


_mod_workingtree.format_registry.register_extra_lazy(
    "breezy.plugins.weave_fmt.workingtree", "WorkingTreeFormat2"
)

serializer.format_registry.register_lazy(
    "4", "breezy.plugins.weave_fmt.xml4", "serializer_v4"
)


def load_tests(loader, basic_tests, pattern):
    testmod_names = [
        "test_bzrdir",
        "test_repository",
        "test_store",
        "test_workingtree",
    ]
    basic_tests.addTest(
        loader.loadTestsFromModuleNames(
            ["{}.{}".format(__name__, tmn) for tmn in testmod_names]
        )
    )
    return basic_tests
